package com.almworks.jira.structure.api;


import com.almworks.jira.structure.api.forest.Forest;
import com.opensymphony.user.User;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Collection;
import java.util.List;

/**
 * <p>A structure is a hierarchical list of issues.
 * <code>Structure</code> interface lets you inspect and change properties of a structure,
 * while {@link Forest} contains the hierarchy itself.</p>
 *
 * <p>You typically get an instance of <code>Structure</code> by calling
 * one of the {@link StructureManager} methods. There are <code>get...</code> methods
 * to retrieve the existing structures, and {@link StructureManager#createStructure}
 * method to create a new structure.</p>
 *
 * <p>To get the <code>Forest</code> associated with the structure, you can use
 * {@link #getForest} method from this class, but you don't have to get an
 * instance of <code>Structure</code> to get the forest - methods
 * {@link StructureManager#getForest} and {@link StructureManager#getForestUpdate}
 * accept the structure ID.</p>
 *
 * <p>To change properties of a structure, you need to get the instance of
 * <code>Structure</code>, call some of the <code>set...</code> methods and then
 * call {@link #saveChanges}. Until you call <code>saveChanges()</code> the modifications
 * are made only to the copy of the structure in memory.</p>
 *
 * <p>To create a new structure, call {@link StructureManager#createStructure} to get
 * an empty <code>Structure</code> instance, call <code>set...</code> methods to set
 * properties (at least structure name must be set) and then call {@link #saveChanges}.
 * When <code>saveChanges</code> is completed, the new Structure has been created and
 * you can use {@link #getId()} to get the new structure's ID.</p>
 *
 * <p>This interface is <strong>not thread-safe</strong>. Every thread must obtain their
 * own copy of the <code>Structure</code> instance. Moreover, every time you need an
 * instance of <code>Structure</code>, you should get it from the <code>StructureManager</code>
 * because the properties might have changed or the structure might have been deleted.</p>
 *
 * @author Igor Sereda
 * @see StructureManager
 * @see Forest
 */
public interface Structure extends Comparable<Structure> {
  // GETTERS

  /**
   * Returns the ID of the structure, or <code>0</code> if the structure has no ID (when it's
   * new and hasn't been saved yet).
   *
   * @return the ID of the structure, or 0 if it doesn't yet have an ID
   */
  long getId();

  /**
   * @return the owner of the structure or null if there's no owner
   */
  @Nullable
  PermissionSubject getOwner();

  /**
   * @return the name of the structure, or null if the name hasn't been set (only possible when the
   *         structure is new)
   */
  @Nullable
  String getName();

  /**
   * @return the description of the structure, or null if the description is not set
   */
  @Nullable
  String getDescription();

  /**
   * @return true if modification of the forest requires <i>Edit Issue</i> permission from the user on the
   *         parent issues of the issues being modified
   * @see <a href="http://wiki.almworks.com/display/structure/Structure+Permissions">Structure Permissions (Structure Documentation)</a>
   */
  boolean isEditRequiresParentIssuePermission();

  /**
   * @return a list of permissions for this structure. User's access level is defined by applying those rules in the order
   *         they are returned from this method. The last matching rule has precedence.
   * @see <a href="http://wiki.almworks.com/display/structure/Structure+Permissions">Structure Permissions (Structure Documentation)</a>
   */
  @NotNull
  List<PermissionRule> getPermissions();

  // MUTATOR METHODS

  /**
   * <p>Sets the ID of the structure. You can only set the ID of a structure instance that's newly created with
   * {@link StructureManager#createStructure} method - this lets you create new structures with a predefined ID.
   * Structure ID must be a positive number, additional limitations on the allowed predefined numbers may be
   * imposed by the implementation.</p>
   *
   * <p>To store the changed information in the database, use {@link #saveChanges}.</p>
   *
   * @param id the ID for the new structure
   * @return this structure
   * @throws IllegalStateException if this is not a new structure
   * @throws IllegalArgumentException if the ID is invalid
   */
  @NotNull
  Structure setId(long id);

  /**
   * <p>Sets the owner of the structure, usually an instance of {@link PermissionSubject.JiraUser}.</p>
   *
   * <p>To store the changed information in the database, use {@link #saveChanges}.</p>
   *
   * @param owner new owner
   * @return this structure
   */
  @NotNull
  Structure setOwner(@Nullable PermissionSubject owner);

  /**
   * <p>Sets the name of the structure. Although it is possible to set name to null, it is an invalid value
   * and <code>saveChanges</code> will throw an error when called, unless a non-null name is set.</p>
   *
   * <p>To store the changed information in the database, use {@link #saveChanges}.</p>
   *
   * @param name structure name
   * @return this structure
   */
  @NotNull
  Structure setName(@Nullable String name);

  /**
   * <p>Sets the description of the structure.</p>
   *
   * <p>To store the changed information in the database, use {@link #saveChanges}.</p>
   *
   * @param description structure description
   * @return this structure
   */
  @NotNull
  Structure setDescription(@Nullable String description);

  /**
   * <p>Changes the security flag that requires the user to have <i>Edit Issue</i> permission on parent
   * issues of the issues being moved, added or deleted from structure.</p>
   *
   * <p>To store the changed information in the database, use {@link #saveChanges}.</p>
   *
   * @param flag if true, <i>Edit Issue</i> permission is required
   * @return this structure
   * @see <a href="http://wiki.almworks.com/display/structure/Structure+Permissions">Structure Permissions (Structure Documentation)</a>
   */
  @NotNull
  Structure setEditRequiresParentIssuePermission(boolean flag);

  /**
   * <p>Sets the permission rules for this structure. Permission rules are checked in the specified order
   * and the last matching rule defines the access level to the structure for the given user.</p>
   *
   * <p>Structure owner and JIRA administrators always have {@link PermissionLevel#ADMIN} access level.</p>
   *
   * <p>To store the changed information in the database, use {@link #saveChanges}.</p>
   *
   * @param permissions a collection of permissions, null means empty list
   * @return this structure
   * @see <a href="http://wiki.almworks.com/display/structure/Structure+Permissions">Structure Permissions (Structure Documentation)</a>
   */
  @NotNull
  Structure setPermissions(@Nullable Collection<? extends PermissionRule> permissions);

  /**
   * <p>Call this method to save the changes made with <code>set...</code> methods and update the database.</p>
   *
   * <p>Before this method is called, all changes are only stored in this instance of Structure and have no other
   * effect. After this method has successfully executed, the changes are stored in the database and applied.</p>
   *
   * <p>The calling user account is passed in the <code>updater</code> parameter. The updater must have
   * {@link PermissionLevel#ADMIN} access level to the structure, and be allowed to use Structure plugin in
   * {@link StructureConfiguration}. Additionally, if this is a new structure being saved, the user must have
   * permissions to create new structures.</p>
   *
   * <p>All security checks are bypassed if <code>overrideSecurity</code> is true.</p>
   *
   * @param updater the user who performs the update, null means anonymous
   * @param overrideSecurity if true, no security checks are made. The <code>user</code> parameter may be used to record change history.
   * @return this structure
   * @throws StructureException if there's a permissions problem, if required fields (name) are not set, or if there
   * are other problems
   */
  @NotNull
  Structure saveChanges(@Nullable User updater, boolean overrideSecurity) throws StructureException;

  // UTILITIES

  /**
   * Calculates access level to this structure for the specified user. If the user is not allowed to use Structure plugin,
   * the result will always be {@link PermissionLevel#NONE}.
   *
   * @param user the user, null means anonymous
   * @return the permissions level
   * @see StructureManager#getStructurePermission
   */
  @NotNull
  PermissionLevel getEffectivePermission(@Nullable User user);

  /**
   * This is a convenience method to get the forest of this structure. More universal method is to use
   * {@link StructureManager#getForestUpdate}.
   *
   * @param user the user requesting the forest
   * @param overrideSecurity if true, no security checks are performed
   * @return the writable copy of the Forest for this structure
   * @throws StructureException if there are any problems (permissions problem, no such structure, and others)
   */
  @NotNull
  Forest getForest(@Nullable User user, boolean overrideSecurity) throws StructureException;
}
