package com.almworks.jira.structure.api;

import com.opensymphony.user.User;
import com.atlassian.jira.jql.builder.JqlClauseBuilder;
import com.atlassian.jira.project.Project;
import com.atlassian.query.Query;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.List;

/**
 * <p><code>StructureConfiguration</code> provides access to the configuration parameters of the Structure plugin.</p>
 *
 * <p>Typically, only JIRA administrators have write access to the plugin configuration, so before changing
 * anything you should check if the user has enough privileges - no checking is done by this component.</p>
 *
 * <p>This service also provides some utility methods that are based on the Structure configuration.</p>
 *
 * <p>All methods in this component are thread-safe.</p>
 */
public interface StructureConfiguration {
  /**
   * @return true if Structure plugin is enabled for all projects
   * @see <a href="http://wiki.almworks.com/display/structure/Selecting+Structure-Enabled+Projects">Selecting Structure-Enabled Projects</a>
   */
  boolean isEnabledForAllProjects();

  /**
   * This method returns a list of projects picked by the administrator for the Structure plugin. It
   * returns a list of picked projects even if "enable for all projects" is turned on - so for the purposes
   * of checking whether Structure is available for a project, you need to use {@link #getCurrentlyEnabledProjects()}.
   *
   * @return a list of projects picked by JIRA administrator on the configuration page
   * @see <a href="http://wiki.almworks.com/display/structure/Selecting+Structure-Enabled+Projects">Selecting Structure-Enabled Projects</a>
   */
  @NotNull
  List<Project> getPickedProjects();

  /**
   * @return a list of projects that are allowed to use Structure (if Structure is enabled for all projects,
   * returns all projects)
   */
  @NotNull
  List<Project> getCurrentlyEnabledProjects();

  /**
   * Checks if a given project is enabled for Structure
   *
   * @param project project to check
   * @return true if the project is not null and available for Structure
   * @see #getCurrentlyEnabledProjects()
   */
  boolean isProjectEnabled(@Nullable Project project);

  /**
   * @return true if Structure plugin is enabled for all users (that have access to the enabled projects)
   * @see <a href="http://wiki.almworks.com/display/structure/Restricting+User+Access+to+Structure">Restricting User Access to Structure</a>
   */
  boolean isEnabledForAnyone();

  /**
   * Used to get the list of groups (possibly other permission subjects) that the plugin is enabled for.
   * Note that if the plugin is enabled for all users, this method would still return the groups that
   * an administrator has previously picked, but those groups are not used to check the permissions.
   *
   * @return a list of subjects that the Structure plugin is enabled for (when not enabled for anyone)
   * @see PermissionSubject.JiraGroup
   * @see <a href="http://wiki.almworks.com/display/structure/Restricting+User+Access+to+Structure">Restricting User Access to Structure</a>
   */
  @NotNull
  List<PermissionSubject> getEnabledPermissionSubjects();

  /**
   * @return true if creating new structures is enabled for all users (that have access to Structure plugin at all)
   * @see <a href="http://wiki.almworks.com/display/structure/Changing+Permission+to+Create+New+Structures">Changing Permission to Create New Structures</a>
   */
  boolean isCreateEnabledForAnyone();

  /**
   * Used to get the list of groups (possibly other permission subjects) that are allowed to create
   * new structures.
   * Note that if structure creation is enabled for all users, this method would still return the groups that
   * an administrator has previously picked, but those groups are not used to check the permissions.
   *
   * @return a list of subjects that are allowed to create new structures (when not enabled for anyone)
   * @see PermissionSubject.JiraGroup
   * @see <a href="http://wiki.almworks.com/display/structure/Changing+Permission+to+Create+New+Structures">Changing Permission to Create New Structures</a>
   */
  @NotNull
  List<PermissionSubject> getCreatorPermissionSubjects();

  /**
   * Changes whether the plugin is enabled for all projects.
   *
   * @param enabled if true, Structure plugin will be enabled for all projects
   * @see <a href="http://wiki.almworks.com/display/structure/Selecting+Structure-Enabled+Projects">Selecting Structure-Enabled Projects</a>
   */
  void setEnabledForAllProjects(boolean enabled);

  /**
   * Changes projects that Structure plugin is enabled for. The setting is effective only if
   * {@link #isEnabledForAllProjects} returns false.
   *
   * @param keyList a comma-delimited list of project keys
   * @see <a href="http://wiki.almworks.com/display/structure/Selecting+Structure-Enabled+Projects">Selecting Structure-Enabled Projects</a>
   */
  void setPickedProjectKeys(@Nullable String keyList);

  /**
   * Changes whether the plugin is available for all users.
   *
   * @param enabled if true, Structure plugin will be available to all users that have access to at least one structure-enabled project.
   * @see <a href="http://wiki.almworks.com/display/structure/Restricting+User+Access+to+Structure">Restricting User Access to Structure</a>
   */
  void setEnabledForAnyone(boolean enabled);

  /**
   * Changes the list of the users / groups that the Structure plugin is enabled for. When Structure
   * is configured to be enabled for anyone, this method has no effect.
   *
   * @param subjects comma-delimited list of string-encoded permission subjects
   * @see com.almworks.jira.structure.api.PermissionSubject#toEncodedString()
   * @see <a href="http://wiki.almworks.com/display/structure/Restricting+User+Access+to+Structure">Restricting User Access to Structure</a>
   */
  void setEnabledPermissionSubjectsEncoded(@Nullable String subjects);

  /**
   * Changes whether the creation of new structures is available for all users.
   *
   * @param enabled if true, anyone who has access to Structure can create new structures.
   * @see <a href="http://wiki.almworks.com/display/structure/Changing+Permission+to+Create+New+Structures">Changing Permission to Create New Structures</a>
   */
  void setCreateEnabledForAnyone(boolean enabled);

  /**
   * Changes the list of the users / groups that are allowed to create new structures. When Structure is configured
   * to allow anyone to create new structures, this setting has no effect.
   *
   * @param subjects comma-delimited list of string-encoded permission subjects
   * @see com.almworks.jira.structure.api.PermissionSubject#toEncodedString()
   * @see <a href="http://wiki.almworks.com/display/structure/Changing+Permission+to+Create+New+Structures">Changing Permission to Create New Structures</a>
   */
  void setCreatorPermissionSubjectsEncoded(@Nullable String subjects);

  /**
   * <p>Adds to the JQL builder a condition that limits the result set to the projects enabled for Structure.</p>
   * <p>This method adds JQL constraint <code>project IN (....)</code> in case Structure is allowed only for
   * specific projects and does nothing if Structure is allowed for all projects. The condition is added at the
   * current level in the clause builder.</p>
   *
   * @param builder a clause builder to be modified; if null, a new clause builder is created
   * @return the builder that was passed as the parameter or the newly created builder
   */
  @NotNull
  JqlClauseBuilder addConfigurationScopeClause(@Nullable JqlClauseBuilder builder);

  /**
   * Utility method that returns a JQL query that selects all issues in the projects that are available for Structure.
   * This is a convenience method that uses {@link #addConfigurationScopeClause}.
   *
   * @return a query that selects issues enabled for Structure, or null if all issues are enabled
   */
  @Nullable
  Query getConfigurationScopeQuery();

  /**
   * Checks if Structure plugin is available for the specified user.
   *
   * @param user the user, null means anonymous
   * @return true if the user can use Structure
   * @see <a href="http://wiki.almworks.com/display/structure/Who+Has+Access+to+the+Structure">Who Has Access to the Structure</a>
   */
  boolean isStructureAvailable(@Nullable User user);

  /**
   * Checks if the user is allowed to create new structures.
   *
   * @param user the user, null means anonymous
   * @return true if the user can create new structures
   * @see <a href="http://wiki.almworks.com/display/structure/Who+Has+Access+to+the+Structure">Who Has Access to the Structure</a>
   */
  boolean isStructureCreationAllowed(@Nullable User user);
}
