package com.almworks.jira.structure.api;

import com.almworks.integers.IntList;
import com.almworks.integers.LongList;
import com.almworks.jira.structure.api.event.StructureListener;
import com.almworks.jira.structure.api.forest.Forest;
import com.almworks.jira.structure.api.forest.ForestAccessor;
import com.almworks.jira.structure.api.forest.ForestTransaction;
import com.almworks.jira.structure.api.forest.ForestUpdate;
import com.opensymphony.user.User;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.List;

/**
 * <p>
 * <code>StructureManager</code> is the main component used to manage structures (as entities) and the forests (the contents of structures).
 * </p>
 * <p>
 * A {@link Structure} is a named entity that contains a hierarchical list of issues ({@link Forest}). Structure has
 * properties like name, description and permissions. One subset of <code>StructureManager</code> methods
 * allows creating/removing structures and changing those properties.
 * </p><p>
 * Another set of methods is dedicated to working with <code>Forest</code> instances - changing the contents of
 * a structure is different from changing structure's properties.
 * </p>
 * <p>Unless otherwise stated, all methods are thread-safe.
 * Note that the methods of the {@link Structure} interface are not thread-safe - you need
 * to retrieve a separate instance in every thread where you need it. </p>
 * <p>
 * Unlike changing a structure, updating a forest is a transactional operation - see {@link #updateForest}.
 * </p>
 * <p>
 * All methods that return <code>Structure</code> instances return new independent instances every time they are
 * called.
 * </p>
 * <h1>Examples</h1>
 * <p>
 * Creating a structure:
 * </p>
 * <pre>
 *   Structure newStructure =
 *     structureManager.createStructure().setName("structureName").saveChanges(user, false);
 * </pre>
 * <p>
 * Changing a structure:
 * </p>
 * <pre>
 *   List&lt;Structure&gt; structures =
 *     structureManager.getStructuresByName("hideme", user, ADMIN, false);
 *   for (Structure s : structures) {
 *     s.setPermissions(Collections.emptyList()).saveChanges(user, false);
 *   }
 * </pre>
 * <p>
 * Updating a forest:
 * </p>
 * <pre>
 *   Structure structure = structureManager.getStructure(structureId, user, EDIT, false);
 *   structureManager.updateForest(user, false, new ForestTransaction&lt;Object&gt;() {
 *      Object transaction(ForestAccessor accessor) throws StructureException {
 *        accessor.moveSubtree(structureId, issueId, under, after);
 *        return null;
 *      }
 *   });
 * </pre>
 *
 * @author Igor Sereda
 * @see Structure
 * @see Forest
 */
public interface StructureManager {
  /**
   * <p>Retrieves a structure given structure ID. Throws exception if the structure cannot be found.</p>
   * <p>Additional parameters allow to specify the required permission level for a user.</p>
   * <p>Structure with ID 1 is the default global structure - see {@link com.almworks.jira.structure.util.StructureUtil#GLOBAL_STRUCTURE_ID}.</p>
   * <p>Unless <code>overrideSecurity</code> is true the user is also checked for being allowed to use Structure plugin.</p>
   *
   * @param structureId the ID of the structure, nullable for convenience (passing null would guarantee to throw StructureException)
   * @param user the user to check permissions for, <code>null</code> for anonymous. If <code>overrideSecurity</code> is true, no checking is made.
   * @param requiredLevel when checking for user's permission, require that the user has at least the specified permission.
   * Passing null or {@link PermissionLevel#NONE} effectively disables the structure permissions check, but the user will be checked for
   * being allowed to use Structure plugin.
   * @param overrideSecurity if true, no permission checks are made whatsoever
   * @return an instance of a structure, not null
   * @throws StructureException if structure is not found, or the user is not allowed to use Structure, or the user does not have the required
   * access level to this structure, or if any other problem is encountered
   */
  @NotNull
  Structure getStructure(@Nullable Long structureId, @Nullable User user, @Nullable PermissionLevel requiredLevel,
    boolean overrideSecurity) throws StructureException;

  /**
   * <p>Retrieves a list of all structures that are visible to the user.</p>
   * <p>
   * The method scans all existing structures and looks for those that the user has the required permission for.
   * </p>
   *
   * @param user the user to check permissions for, <code>null</code> for anonymous. If <code>overrideSecurity</code> is true, no checking is made.
   * @param requiredLevel when checking for user's permission, require that the user has at least the specified permission for every
   * structure in the returned list.
   * Passing null or {@link PermissionLevel#NONE} effectively disables the structure permissions check, but the user will be checked for
   * being allowed to use Structure plugin.
   * @param overrideSecurity if true, no permission checks are made
   * @return a list of structures available to the user, not null
   */
  @NotNull
  List<Structure> getAllStructures(@Nullable User user, @Nullable PermissionLevel requiredLevel,
    boolean overrideSecurity);

  /**
   * <p>Convenience method to search for structures with the specified name. Using this method is more
   * efficient than retrieving all structures and filtering.</p>
   * <p>Structures are not required to have unique names, therefore this method returns a list.</p>
   *
   * @param name the name of the structure sought, <strong>case-insensitive</strong> and <strong>trim-insensitive</strong> (leading
   * and trailing whitespace doesn't matter)
   * @param user the user to check permissions for, <code>null</code> for anonymous. If <code>overrideSecurity</code> is true, no checking is made.
   * @param requiredLevel when checking for user's permission, require that the user has at least the specified permission for every
   * structure in the returned list.
   * Passing null or {@link com.almworks.jira.structure.api.PermissionLevel#NONE} effectively disables the structure permissions check, but the user will be checked for
   * being allowed to use Structure plugin.
   * @param overrideSecurity if true, no permission checks are made
   * @return a list of the matching structures
   */
  @NotNull
  List<Structure> getStructuresByName(@Nullable String name, @Nullable User user,
    @Nullable PermissionLevel requiredLevel, boolean overrideSecurity);

  /**
   * <p>Retrieves all structures that contain the specified issue. This is functionally the same as checking every forest
   * manually, but a lot faster.</p>
   * <p>
   * Note that forest changes may happen asynchronously.
   * The result of this call is guaranteed to be consistent for a certain point in time.
   * </p>
   *
   * @param issue the ID of the issue
   * @param user the user to check permissions for, <code>null</code> for anonymous. If <code>overrideSecurity</code> is true, no checking is made.
   * @param requiredLevel when checking for user's permission, require that the user has at least the specified permission for every
   * structure in the returned list.
   * Passing null or {@link com.almworks.jira.structure.api.PermissionLevel#NONE} effectively disables the structure permissions check, but the user will be checked for
   * being allowed to use Structure plugin.
   * @param overrideSecurity if true, no permission checks are made
   * @return a list of structures
   */
  @NotNull
  List<Structure> getStructuresWithIssue(@Nullable Long issue, @Nullable User user,
    @Nullable PermissionLevel requiredLevel, boolean overrideSecurity);

  /**
   * Used to show the user the structures that have been recently updated.
   *
   * @param user the user to check permissions for, <code>null</code> for anonymous - the structures will be filtered for this user
   * @param requiredLevel the required access level for all structures in the returned list.
   * Passing null or {@link com.almworks.jira.structure.api.PermissionLevel#NONE} effectively disables the structure permissions check, but the user will be checked for
   * being allowed to use Structure plugin.
   * @param maxCount maximum number of elements in the result
   * @return a list of structures, in order from the most recently updated (by anyone)
   */
  @NotNull
  List<Structure> getRecentlyUpdatedStructures(@Nullable User user, @Nullable PermissionLevel requiredLevel,
    int maxCount);

  /**
   * Used to check if the user has at least one structure with the specified access level.
   *
   * @param user the user to check permissions for, <code>null</code> for anonymous
   * @param requiredLevel the minimum access level for the specified user.
   * Passing null or {@link com.almworks.jira.structure.api.PermissionLevel#NONE} effectively
   * makes this method to check if there are any structures at all and if the user has access to the Structure plugin.
   * @return true if the user has access to the Structure plugin and there exists at least one structure that the user
   *         has access to with the specified permission level
   */
  boolean hasStructuresForUser(@Nullable User user, @Nullable PermissionLevel requiredLevel);

  /**
   * Used to check if Structure plugin is used. Functionally it is the same as getting all structures and checking
   * if their forests are not empty, but much faster.
   *
   * @return true if there's at least one structure that contains at least one issue
   */
  boolean hasStructuresWithIssues();

  /**
   * Checks if a specific issue belongs to a specific structure.
   *
   * @param issueId the ID of the issue
   * @param structureId the ID of the structure
   * @return true if the the issue is in the structure
   */
  boolean isIssueInStructure(@Nullable Long issueId, @Nullable Long structureId);

  /**
   * <p>Finds the first issue among the passed list that belongs to any structure at all. Used to see if issues
   * are in the structure.</p>
   *
   * @param issueIds a list with issue IDs
   * @return index of the first issue in the list that is added to at least one structure, or -1 if none of the
   *         issues are in the structure or the list is null
   */
  int indexOfIssueInAnyStructure(@Nullable LongList issueIds);

  /**
   * Calculates access level to a structure for the specified user. If the user is not allowed to use Structure plugin,
   * the result will always be NONE.
   *
   * @param structureId the ID of the structure checked
   * @param user the user to check permissions for, <code>null</code> for anonymous
   * @return the calculated access level
   */
  @NotNull
  PermissionLevel getStructurePermission(Long structureId, User user);

  /**
   * Used to check whether the specified structure exists and visible to the user with the required permission level.
   *
   * @param structureId the ID of the structure
   * @param user the user to check permissions for, <code>null</code> for anonymous. If <code>overrideSecurity</code> is true, no user check is made.
   * @param requiredLevel when checking for user's permission, require that the user has at least the specified permission level for the structure.
   * Passing null or {@link com.almworks.jira.structure.api.PermissionLevel#NONE} effectively disables the structure permissions check, but the user will be checked for
   * being allowed to use Structure plugin.
   * @param overrideSecurity if true, no permission checks are made
   * @return true if the specified structure exists and the user has access to it with the specified access level
   */
  boolean isAccessible(@Nullable Long structureId, @Nullable User user, @Nullable PermissionLevel requiredLevel,
    boolean overrideSecurity);

  /**
   * Creates an empty new structure. The structure returned is not yet persisted - you need to call the required
   * setter methods and then call {@link Structure#saveChanges(com.opensymphony.user.User, boolean)}
   * to write the new structure to the database and be able to create the forest.
   *
   * @return the instance of the new structure, not persisted to the database
   * @see Structure#saveChanges
   */
  @NotNull
  Structure createStructure();

  /**
   * Copies the structure and the forest it contains into a new structure.
   *
   * @param structureId the ID of the structure
   * @param newOwner the user who will be the owner of the new structure, if null - keep the old owner
   * @param copyPermissions if true, the permission list will be copied - this requires ADMIN permission level to the
   * original structure, if the user does not have it, StructureException is thrown
   * @param user the user who performs the operation, null means anonymous. This user will be checked for having
   * all required permissions (using Structure plugin, creating structures, ADMIN access to the structure being copied)
   * @param overrideSecurity if true, no permission checks are made
   * @return the freshly created structure. You can use {@link com.almworks.jira.structure.api.Structure#getId()}
   *         to learn its ID
   * @throws StructureException if the user has not enough permissions or if any problem happens
   */
  @NotNull
  Structure copyStructure(@Nullable Long structureId, @Nullable User newOwner, boolean copyPermissions,
    @Nullable User user, boolean overrideSecurity) throws StructureException;

  /**
   * Deletes the specified structure and its content.
   *
   * @param id the ID of the structure
   * @param user the user performing the operation, null means anonymous
   * @param overrideSecurity if true, no permission checks are made
   * @throws StructureException if there's problem deleting structure
   */
  void deleteStructure(@Nullable Long id, @Nullable User user, boolean overrideSecurity) throws StructureException;

  /**
   * <p>Deletes all structures completely, effectively emptying the database.</p>
   * <p>To execute this operation, the user has to have ADMINISTER permissions in JIRA - it is not enough to have
   * ADMIN permission levels on the structures.</p>
   *
   * @param user the user performing the operation, null means anonymous (and guarantees that exception is thrown)
   * @throws StructureException if there's a permissions problem or any other problem
   */
  void deleteAllStructures(@Nullable User user) throws StructureException;

  /**
   * <p>Creates an instance of {@link ForestUpdate} for the given structure - a difference report between the
   * version of the forest known by the client code and the latest version of the forest.</p>
   * <p>This is a universal method to retrieve a forest. To get the full forest, you need to pass <code>0</code>
   * as the version and expect to receive {@link ForestUpdate.Full} in reply. However, there are convenience
   * methods {@link #getForest} and {@link Structure#getForest}.</p>
   * <p>
   *   The returned result has one of the three possible types, which the client code can check using <code>instanceof</code>:
   * </p>
   * <ul>
   *   <li>{@link ForestUpdate.Empty} means that the client has the most recent version of the Forest.</li>
   *   <li>{@link ForestUpdate.Incremental} means that there are several operations to be performed on the
   *   forest to bring it to the last version.</li>
   *   <li>{@link ForestUpdate.Full} means that the reply contains the whole forest.</li>
   * </ul>
   * <p>
   *   If <code>root</code> parameter is not null and not <code>0L</code>, the resulting forest is filtered
   *   for this root issue - that is, only the parent issues and the sub-issues (of any level) will be included.
   *   Other "unrelated" issues will not be included in the forest.
   * </p>
   * <p>
   *   If incremental response is requested (with <code>version</code> parameter) and <code>root</code> is also
   *   specified, then the response will be incremental only if the diff contains the operations "visible" in the
   *   reduced forest.
   * </p>
   * <p>
   *   The issues returned by this method are checked to be visible to the specified user. The user is also checked
   *   to have permission to use Structure and VIEW permission level on the specified structure. In case some issues
   *   in the forest are not visible to the user, the forest is reduced to contain only the visible issues -
   *   see {@link Forest#filterSoft}. This also applies to incremental updates - if an incremental update would
   *   concern an issue not visible to the user, it will fall back to a full update.
   * </p>
   *
   * @param structureId the ID of the structure
   * @param version the version known to the calling code, or null or 0L if no version is known or full update is required
   * @param root an ID of a "root" issue when only a part of the forest requested, null or 0L otherwise
   * @param user the user performing the operation, null means anonymous
   * @param overrideSecurity if true, no security check will be performed and issues won't be filtered
   * @return forest update to bring the state of the forest known to the client code to the current version
   * @throws StructureException if the structure does not exist or if the user does not have VIEW permissions on it
   */
  @NotNull
  ForestUpdate getForestUpdate(@Nullable Long structureId, @Nullable Long version, @Nullable Long root,
    @Nullable User user, boolean overrideSecurity) throws StructureException;

  /**
   * <p>Retrieves the forest of the given structure. This is a convenience method that uses {@link #getForestUpdate} method.</p>
   * <p>Unless <code>overrideSecurity</code> is <code>true</code>, the returned forest is filtered for the
   * user passed as the parameter.</p>
   * <p>Note that the returned forest is a mutable copy - you can mutate it, but it won't change the forest in the structure.</p>
   *
   * @param structureId the ID of the structure to retrieve
   * @param user the user performing the operation, null means anonymous
   * @param overrideSecurity if true, no security check will be performed and issues won't be filtered
   * @return structure's forest
   * @throws StructureException if the structure does not exist or if the user does not have VIEW permissions on it
   */
  @NotNull
  Forest getForest(@Nullable Long structureId, @Nullable User user, boolean overrideSecurity) throws StructureException;

  /**
   * Retrieves current version of the forest in the specified structure.
   *
   * @param structureId the ID of the structure
   * @param user the user performing the operation, null means anonymous
   * @param overrideSecurity if true, no security check will be performed and issues won't be filtered
   * @return the version of the forest, or <code>0</code> if the structure does not exist or the user does not have access to it
   */
  long getForestVersion(@Nullable Long structureId, @Nullable User user, boolean overrideSecurity);

  /**
   * <p>Used to quickly retrieve the path to an issue in the given structure. The path contains all ancestor issues
   * of the specified issue (parent, grand-parent and so on) in the structure's forest, and ends with the issue itself.</p>
   *
   * <p>If the issue is not in the structure, or if structure with the given ID does not exist, returns an empty list.</p>
   *
   * <p>This method does not do any permission checks.</p>
   *
   * @param structureId the ID of the structure
   * @param issueId the ID of the issue
   * @return the path from top-level issue to the specified issue in the specified structure's forest
   * @see Forest#getPath(long)
   */
  @NotNull
  LongList getIssuePathIgnorePermissions(@Nullable Long structureId, @Nullable Long issueId);

  /**
   * <p>Performs an atomic update operation on forests.</p>
   *
   * <p>This method is the only way to update a hierarchy stored in the structure. You should enclose the code
   * that makes the updates in a {@link ForestTransaction} and pass to this method. The manipulation of the
   * forests is done via primitive methods defined in {@link ForestAccessor},
   * and the accessor is passed to the <code>ForestTransaction</code> as a parameter. </p>
   *
   * <p>
   *   <strong>It is very important to implement <code>ForestTransaction</code> correctly.</strong> The transaction
   *   may be executed under one or more write locks, preventing other transactions from changing the forests
   *   concurrently (readers still have access) - so the amount of code should be minimal and the execution time
   *   should be predictable and small.
   * </p>
   * <p>
   *   Putting the following in a <code>ForestTransaction</code> is a <strong>bad idea:</strong>
   * </p>
   * <ul>
   *   <li>Accessing <code>synchronized</code> blocks.</li>
   *   <li>Calling listeners or other code that you don't control.</li>
   *   <li>Running a long calculation that can be done outside the transaction.</li>
   *   <li>Accessing file system or network.</li>
   *   <li>Blocking execution for whatever reason.</li>
   * </ul>
   * <p>
   *   The update is atomic - it either is fully successful or fully failed. If an exception is thrown from the
   *   <code>ForestTransaction</code>'s {@link ForestTransaction#transaction} method, the transaction rolls back
   *   and no changes are applied. If no exception is thrown, the transaction is considered completed and all changes
   *   are committed into the database.
   * </p>
   * <p>
   *   Note that you can wrap the calls to <code>ForestAccessor</code>'s method into try-catch block to prevent
   *   an exception escaping the transaction and causing the rollback. In that case, the transaction will be
   *   considered successful, but the result of the failed operation is not defined - it may be only half-completed
   *   (for example, with {@link ForestAccessor#mergeForest} operation).
   * </p>
   * <p>
   *   The update is serialized - full reader isolation is provided. Someone accessing forests being changed using
   *   <code>StructureManager</code>'s {@link #getForestUpdate} method will not see the results of a transaction until
   *   it's successfully finished and there's no chance to see inconsistent data.
   *   Because of that, <code>ForestAccessor</code> has method {@link ForestAccessor#getForest} that retrieves the
   *   forest as seen at the "current moment" in the transaction.
   * </p>
   *
   * @param user the user that performs the update, null means anonymous. The user will be checked for access to the
   * Structure plugin and having {@link PermissionLevel#EDIT} permission level on the structures that the transaction
   * updates.
   * @param overrideSecurity if true, no security check will be performed and issues won't be filtered
   * @param operation the update transaction, typically an anonymous class. Null is acceptable and does nothing (returns null).
   * @param <T> type of the transaction result. If transaction does not need to return a result, use <code>Object</code> or <code>Void</code>.
   * @return the value that has been returned by <code>operation</code>
   * @throws StructureException if anything goes wrong or if exception is thrown by the operation
   */
  <T> T updateForest(@Nullable User user, boolean overrideSecurity, @Nullable ForestTransaction<T> operation) throws StructureException;

  /**
   * Adds a structure listener. Whenever a structure is changed, the listeners will be called synchronously right after
   * transaction completes.
   *
   * @param listener the listener
   */
  void addListener(@Nullable StructureListener listener);

  /**
   * Removes the listener previously added via {@link #addListener}.
   *
   * @param listener listener to be removed
   */
  void removeListener(@Nullable StructureListener listener);

  /**
   * Creates an empty instance of <code>Forest</code>.
   *
   * @return new empty forest
   * @see Forest
   */
  @NotNull
  Forest createForest();

  /**
   * <p>Creates an instance of <code>Forest</code> that has the specified list of issues and depths.</p>
   *
   * <p>The passed lists must satisfy the invariant conditions listed in {@link Forest}, otherwise the results
   * are undefined.</p>
   *
   * @param issues issues list
   * @param depths depths list
   * @return new forest with the specified issues at the specified depths
   * @see Forest the new forest
   */
  @NotNull
  Forest createForest(@NotNull LongList issues, @NotNull IntList depths);
}
