package com.almworks.jira.structure.api;

import com.almworks.jira.structure.api.backup.StructureBackupManager;
import com.almworks.jira.structure.api.event.IssueEventBridge;
import com.almworks.jira.structure.api.job.StructureJobManager;
import com.almworks.jira.structure.api.progress.ProgressCalculator;
import com.almworks.jira.structure.api.sync.StructureSyncManager;
import com.almworks.jira.structure.api.sync.SynchronizerUndoRecorder;
import com.atlassian.jira.issue.IssueManager;
import com.atlassian.jira.security.JiraAuthenticationContext;
import org.jetbrains.annotations.NotNull;

/**
 * <p>
 * StructureServices is a convenience interface that can be used to retrieve all public services
 * of the Structure plugin. You can also get those services from ComponentAccessor or by declaring them
 * in a constructor of your component and have them injected.
 * </p>
 * <p>
 * The convenience is in reducing the number of parameters you declare in the constructor of your
 * component.
 * </p>
 * <p>
 * Note that this service returns the original instances of those services, unlike when the proxies for the
 * services are injected.
 * </p>
 * <p>
 * This class also provides some frequently-used Atlassian interfaces.
 * </p>
 *
 * @author Igor Sereda
 */
public interface StructureServices {
  /**
   * @return component that's used to work with structures and forests
   */
  @NotNull
  StructureManager getStructureManager();

  /**
   * @return component that's used to manage structure configuration
   */
  @NotNull
  StructureConfiguration getStructureConfiguration();

  /**
   * @return component that's used to calculate issue progress
   */
  @NotNull
  ProgressCalculator getProgressCalculator();

  /**
   * @return component that's used to work with synchronizers
   */
  @NotNull
  StructureSyncManager getSyncManager();

  /**
   * @return component that's used to back up and restore structure
   */
  @NotNull
  StructureBackupManager getBackupManager();

  /**
   * @return component that's used to schedule asynchronous jobs
   */
  @NotNull
  StructureJobManager getJobManager();

  /**
   * @return component that's used to listen and report issue events
   */
  @NotNull
  IssueEventBridge getIssueEventBridge();

  /**
   * @return component that's used by the synchronizers to store undo information
   */
  @NotNull
  SynchronizerUndoRecorder getSynchronizerUndoRecorder();

  /**
   * @return JIRA component that's used to get the current user
   */
  @NotNull
  JiraAuthenticationContext getAuthenticationContext();

  /**
   * @return JIRA component that's used to work with issues
   */
  @NotNull
  IssueManager getIssueManager();
}
