package com.almworks.jira.structure.api.event;

import com.almworks.integers.LongList;
import com.almworks.jira.structure.api.StructureManager;
import org.jetbrains.annotations.NotNull;

import java.util.Map;

/**
 * <p><code>StructureListener</code> allows you to listen to changes in structures.</p>
 *
 * <p>Whenever structure change transaction, executed with {@link StructureManager#updateForest},
 * is finished, all subscribed listeners are called <strong>in the same thread</strong>. This means that
 * the listener must execute as quickly as possible and should avoid taking locks. Otherwise
 * the listener may block the whole Structure database, because write transactions may hold
 * write lock for the whole database or for the changed structures.</p>
 *
 * @author Igor Sereda
 */
public interface StructureListener {
  /**
   * Called at the end of transaction that changes structure forests. The passed parameter
   * allows the listener to analyze and act on the changes.
   *
   * @param affectedMap the map (structure ID => changes)
   * @see StructureChanges
   */
  void onStructureChanged(@NotNull Map<Long, StructureChanges> affectedMap);

  /**
   * <p><code>StructureChanges</code> class represents aggregate information about changes
   * that happened to a single structure. It contains the list of the changed issues,
   * the list of direct parents of the affected issues, the list of all ancestors of the affected issues
   * and forest version information.</p>
   *
   * <p>A listener often needs to recalculate some information that is based on the hirerarchy.
   * This class contains enough information to build a list of issues that need recalculation.</p>
   */
  class StructureChanges {
    private final LongList myAffectedIssuesSorted;
    private final LongList myParentsSorted;
    private final LongList myAncestorsSorted;
    private final long myVersionBefore;
    private final long myVersionAfter;

    /**
     * Constructs the changes record.
     *
     * @param versionBefore version of the forest that was present before the change
     * @param versionAfter version of the forest after the change
     * @param affectedIssuesSorted sorted list of affected issues with unique values
     * @param parentsSorted sorted list of parents of the affected issues
     * @param ancestorsSorted sorted list of all ancestors of the affected issues
     */
    public StructureChanges(long versionBefore, long versionAfter, @NotNull LongList affectedIssuesSorted,
      @NotNull LongList parentsSorted, @NotNull LongList ancestorsSorted)
    {
      assert affectedIssuesSorted.isUniqueSorted() : affectedIssuesSorted;
      assert parentsSorted.isUniqueSorted() : parentsSorted;
      assert ancestorsSorted.isUniqueSorted() : ancestorsSorted;
      myVersionBefore = versionBefore;
      myVersionAfter = versionAfter;
      myAffectedIssuesSorted = affectedIssuesSorted;
      myParentsSorted = parentsSorted;
      myAncestorsSorted = ancestorsSorted;
    }

    /**
     * @return sorted list of IDs of the issues affected by the change (moved, added or deleted from structure)
     */
    public LongList getAffectedIssuesSorted() {
      return myAffectedIssuesSorted;
    }

    /**
     * @return sorted list of IDs of the parent issues of the affected issues
     */
    public LongList getParentsSorted() {
      return myParentsSorted;
    }

    /**
     * @return sorted list of IDs of the ancestor issues (parent, grand-parent and so on) of the affected issues
     */
    public LongList getAncestorsSorted() {
      return myAncestorsSorted;
    }

    /**
     * @return version of the forest before the change
     */
    public long getVersionBefore() {
      return myVersionBefore;
    }

    /**
     * @return version of the forest after the change
     */
    public long getVersionAfter() {
      return myVersionAfter;
    }
  }
}
