package com.almworks.jira.structure.api.forest;

import com.almworks.integers.*;
import com.almworks.jira.structure.api.StructureException;
import com.almworks.jira.structure.api.StructureManager;
import com.almworks.jira.structure.util.La;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * <p><code>Forest</code> interface provides access to a data structure that stores issue hierarchy
 * (the hierarchy is a sequence of tree structures, a forest).
 * <code>Forest</code> lets you read and modify the data stored in the forest data structure.</p>
 *
 * <p>Difference between {@link com.almworks.jira.structure.api.Structure} and <code>Forest</code>
 * interfaces: <code>Structure</code> exposes properties of a structure such as name, and also lets you
 * modify those properties and save them back to the database; while <code>Forest</code> is the hierarchy
 * itself, it's just a data structure. Any modifications to the <code>Forest</code> are only applied to
 * that instance in memory.</p>
 *
 * <p>
 *   Forest of issues is represented as a sequence of pairs <code>(long issue, int depth)</code>. The issues
 *   are identified by their issue ID, and the position in the hierarchy is identified by the position
 *   in the sequence and the associated depth. For example, the following hierarchy:
 *   <pre>
 *     A
 *        A1
 *        A2
 *     B
 *        B1
 *            B1X
 *   </pre>
 *   is represented in the forest as the following sequence:
 *   <pre>
 *     (A, 0), (A1, 1), (A2, 1), (B, 0), (B1, 1), (B1X, 2)
 *   </pre>
 * </p>
 *
 * <p>
 *   <code>Forest</code> interface provides indexed access to that representation of the forest.
 *   That is, you can read issue IDs and depths given that you know the index of the pair in the
 *   sequence. There are also a number of utility methods that allow searching and browsing the
 *   forest.
 * </p>
 *
 * <h1>Invariants</h1>
 * <p>
 *  Because <code>depth</code> associated with an issue is an integer number, there are certain
 *  invariants about the depth that must hold true, otherwise the data does not make sense. Additionally,
 *  an issue may appear no more than once in the forest. Make sure
 *  you don't violate those invariants when creating new forest instances:
 * </p>
 * <ul>
 *   <li>The depth of the first element in the forest must be <code>0</code>.</li>
 *   <li>Given that the depth of an element is <code>D</code>,
 *   the depth of the following element must be within range <code>[0 .. D+1]</code>.
 *   </li>
 *   <li>If an issue is present in the forest at index <code>N</code>, it must not be present
 *   at any other index.</li>
 *   <li>Issue ID may not be <code>0</code> - this value represents "missing" or "no value" in various methods of this interface.</li>
 * </ul>
 * <p>
 * When Java assertions are enabled, these invariants are checked all the time. Due to performance impact,
 * the invariants are not checked and assumed to hold true when running with assertions disabled.
 * The result of creating and using a forest that violates these invariants is undefined.
 * </p>
 *
 * <h1>Under-After Coordinates</h1>
 * <p>
 *   When a specific position in the forest needs to be defined for mutating operations such as move or insert, <code>under-after</code>
 *   coordinates are used. (If operation addresses a specific issue, the issue ID is used instead.)
 * </p>
 * <p>
 *   <code>Under-after</code> coordinates contain <code>under</code> value, which identifies the parent issue of the
 *   specified position (the depth and the subtree). If <code>under</code> is <code>0</code>, the position is at the
 *   top (root issue).
 *</p>
 * <p>The other parameter, <code>after</code> coordinate, defines the location of the inserted issue amongst siblings -
 * the issues at the same depth under the same parent. If <code>after</code> is <code>0</code>, it means that the position
 * should be the first under the specified parent.</p>
 *
 * <p>This class in <strong>not thread-safe</strong>.</p>
 *
 * @author Igor Sereda
 */
public interface Forest {
  // ======================== ACCESSORS - WORK FAST ========================

  /**
   * @return the size of the forest, the number of issues in it, the number is always >= 0
   */
  int size();

  /**
   * @return true if the size of the forest is zero
   */
  boolean isEmpty();

  /**
   * @return a non-modifiable list of issues, in the order they appear in the forest. The size of
   * the list is equal to the value returned by {@link #size()}.
   */
  @NotNull
  LongList getIssues();

  /**
   * @return a non-modifiable list of depths, in the order the issues appear in the forest. The size
   * of the list is equal to the value returned by {@link #size()} and the <code>i</code>-th element
   * of this list corresponds to the issue ID at the <code>i</code>-th position in the list returned
   * by {@link #getIssues()}.
   */
  @NotNull
  IntList getDepths();

  /**
   * Gets the ID of an issue at the specified position in the forest.
   *
   * @param index the index of the forest entry
   * @return issue ID at the specified index
   * @throws IndexOutOfBoundsException if <code>index</code> is not within range <code>[0 .. size() - 1]</code>.
   */
  long getIssue(int index);

  /**
   * Gets the depth of an issue at the specified position in the forest.
   *
   * @param index the index of the forest entry
   * @return depth at the specified index, value is always >= 0 and satisfies the invariants for the {@link Forest}
   * @throws IndexOutOfBoundsException if <code>index</code> is not within range <code>[0 .. size() - 1]</code>.
   */
  int getDepth(int index);


  // ======================== SEARCHERS - WORK SLOWER ========================

  /**
   * <p>Searches for the position of a specific issue in the forest.</p>
   *
   * <p>Execution of this method may take <code>O(size())</code> time, however it may be
   * optimized if the implementation maintains an index. It's better to use this method rather
   * than using <code>forest.getIssues().indexOf(issue)</code> because of the possible optimizations.</p>
   *
   * @param issue the issue ID to search for
   * @return such index of the issue in the forest, or <code>-1</code> if the issue is not found
   */
  int indexOf(long issue);

  /**
   * <p>Can be used to check if the forest contains a specific issue.</p>
   *
   * <p>Execution of this method may take <code>O(size())</code> time.</p>
   *
   * @param issue the issue ID to search for
   * @return true if the forest contains the specified issue
   */
  boolean containsIssue(long issue);

  /**
   * <p>The method looks for the end of a subtree rooted at the specified index.</p>
   *
   * <p>A subtree rooted at index <code>k</code> is a sub-sequence in the forest starting at position <code>k</code>
   * and containing all following elements that have depth <code>d > depth[k]</code>.</p>
   *
   * <p>The result of this method is the <strong>next</strong> index after the last element of this subsequence.
   * More specifically, the result is the first element after <code>[k]</code> that has depth <code>d <= depth[k]</code>.</p>
   *
   * <p>When the subtree ends with the whole forest, the return value is equal to {@link #size()}.</p>
   *
   * <p>If <code>index</code> is a negative number, the returned value is <code>0</code>. If <code>index</code> is greater or equal than
   * <code>size()</code>, the returned value is equal to <code>size()</code>.</p>
   *
   * @param index the index of the root issue of the subtree
   * @return the index of an issue that follows the last issue of the subtree, or the size of the forest in case the subtree is at the end of it
   */
  int getSubtreeEnd(int index);

  /**
   * <p>Searches the forest for the index of a "parent issue". If the issue at the specified index is a root issue (has depth of <code>0</code>),
   * returns -1.</p>
   *
   * <p>If <code>index</code> is negative, returns -1.</p>
   *
   * @param index index of a child issue
   * @return -1 if the issue at <code>index</code> is a root issue, or index <code>k</code> of the parent issue, such as
   * that <code>k = MAX(k in [0, index-1] that has depth[k] == depth[index] - 1)</code>
   * @throws IndexOutOfBoundsException if <code>index</code> is equal or greater than forest size
   */
  int getParentIndex(int index);

  /**
   * Gets the parent issue of the specified issue.
   *
   * @param issue child issue
   * @return parent issue, or <code>0</code> if the child issue is a root issue or not found in the forest
   */
  long getParent(long issue);

  /**
   * Gets the issue that immediately precedes the specified issue in the
   * list of children of the specified issue's parent.
   *
   * @param issue an issue
   * @return immediately preceding sibling (issue that has the same parent and same depth), or 0 if <code>issue</code> is a root issue or not found in the forest or is the first child of its parent issue
   */
  long getPrecedingSibling(long issue);

  /**
   * Gets the index of an issue that immediately precedes the issue at the given index in the
   * list of children of its parent.
   *
   * @param index issue index
   * @return index of immediately preceding sibling, or -1 if issue at <code>index</code> is a root issue or not found in the forest or is the first child of its parent issue
   */
  int getPrecedingSiblingIndex(int index);

  /**
   * Gets the issue that immediately precedes the one with the given index in the
   * list of children of its parent.
   *
   * @param index issue index
   * @return immediately preceding sibling, or 0 if there is none
   */
  long getPrecedingSiblingForIndex(int index);

  /**
   * Returns the array of all issues that come before the given issue in the list of children of its parent,
   * in the same order as they appear in the forest.
   *
   * @param issue an issue
   * @return array of all preceding siblings, empty if <code>issue</code> has none or is not in the forest
   */
  @NotNull
  LongArray getPrecedingSiblings(long issue);

  /**
   * Returns the array of all issues that come before the given issue in the list of children of its parent,
   * in the same order as they appear in the forest.
   *
   * @param index issue index
   * @return array of all preceding siblings, empty if issue at <code>index</code> has none or <code>index</code> is negative
   */
  @NotNull
  LongArray getPrecedingSiblingsForIndex(int index);

  /**
   * <p>Given issue at the specified index, traverse its "path" upwards - that is, look for all
   * parent issues up to the topmost root parent, and return an index of the parent that has the
   * specified depth.</p>
   *
   * <p>If the required depth is equal to the depth of the issue at the specified index, returns <code>index</code>.</p>
   *
   * <p>If <code>index</code> is a negative value, returns -1. If issue at the specified index has less depth than the required value, returns <code>-1</code>.</p>
   *
   * @param index index of an issue
   * @param depth required depth of the [grand-] parent issue
   * @return the index of the [grand-] parent issue on the "path" to the specified issue that has the specified depth, or <code>-1</code> if not found
   * @throws IndexOutOfBoundsException if index is equal or greater than the size of the forest
   */
  int getPathIndexAtDepth(int index, int depth);

  /**
   * <p>Gets the last direct sub-issue of the specified parent issue.</p>
   *
   * <p><strong>Special case</strong>: when <code>parent</code> is <code>0</code>,
   * returns the last root issue in the forest.</p>
   *
   * <p>If the parent issue is not in the forest, or if it does not have child issues, the return value is <code>0</code>.</p>
   *
   * @param parent parent issue
   * @return the last issue among the parent issue's children, or 0 if the parent issue does not have children
   */
  long getLastChild(long parent);

  /**
   * <p>Gets the last direct sub-issue of the specified parent issue.</p>
   *
   * <p><strong>Special case</strong>: when <code>parentIndex</code> is less than zero,
   * returns the last root issue in the forest.</p>
   *
   * <p>If the parent issue does not have child issues, the return value is <code>0</code>.</p>
   *
   * @param parentIndex the index of the parent issue
   * @return the last issue among the parent issue's children, or 0 if the parent issue does not have children
   */
  long getLastChildByIndex(int parentIndex);


  // ======================== COPIERS - create new data structures with data from the forest ========================

  /**
   * <p>Creates an array with all direct sub-issues of the specified issue.</p>
   *
   * <p>The returned array is writable and owned by the calling code.</p>
   *
   * <p>If issue is not in the forest or does not have children, empty array is returned.</p>
   *
   * @param issue the parent issue
   * @return array of all sub-issues of the specified issue that have depth equal to the parent depth + 1
   */
  @NotNull
  LongArray getChildren(long issue);

  /**
   * <p>Creates an array with all direct sub-issues of the issue at the specified index.</p>
   *
   * <p>The returned array is writable and owned by the calling code.</p>
   *
   * <p>If the specified issue does not have children, or if the index is negative, empty array is returned.</p>
   *
   * @param index the index of the parent issue
   * @return array of all sub-issues of the specified issue that have depth equal to the parent depth + 1
   * @throws IndexOutOfBoundsException if the index is greater or equals the size of the forest
   */
  @NotNull
  LongArray getChildrenAtIndex(int index);

  /**
   * @return an array of all root issues in the forest (those that have depth of <code>0</code>)
   */
  @NotNull
  LongArray getRoots();

  /**
   * <p>Returns the path to the specified issue - a sequence of issues that starts with a
   * root issue, ends with the specified issue, and where <code>element[i]</code> is
   * the parent issue of <code>element[i+1]</code>.</p>
   *
   * <p>If <code>issue</code> is not in the forest, returns empty array.</p>
   *
   * <p>The array is modifiable and owned by the calling code.</p>
   *
   * @param issue issue to get the path to
   * @return path to the specified issue, or empty array if the issue is not in the forest
   */
  @NotNull
  LongArray getPath(long issue);

  /**
   * <p>Returns the path to the specified issue without the issue itself. In other words, this
   * is the path to the parent of the specified issue, if there is one.</p>
   *
   * @see #getPath(long)
   * @param index issue index
   * @return path to the specified issue's parent; empty array if the issue is top-level or not in the forest
   */
  @NotNull
  LongArray getParentPathForIndex(int index);

  /**
   * <p>Filters this forest by hiding issues that do not pass the <code>filter</code> condition. The
   * resulting forest contains only the issues that pass the condition.</p>
   *
   * <p>The topology of the resulting forest may differ from the original forest - that is, an issue may
   * have a different parent in the resulting forest. This happens when an issue that has sub-issues is filtered
   * out - in that case, its sub-tree is substituted instead of the parent issue.
   * This is different from {@link #filterSoft} method, which preserves the topology.</p>
   *
   * <p>This forest is not modified by this method. If all issues pass the condition, then this forest
   * is returned as the result. If filtering has taken place, a new forest is returned.</p>
   *
   * <p>The filter method is called once for every issue in the forest, and a truthy result
   * (as defined in {@link La#accepts}) indicates that the issue passes the filter.</p>
   *
   * @param filter filter that returns a truthy value if an issue with given ID is allowed to be present in the resulting forest
   * @return a filtered forest (or this forest if all issues satisfy the filter)
   * @see #filterSoft
   */
  @NotNull
  Forest filter(@Nullable La<Long, ?> filter);

  /**
   * <p>Filters this forest by excluding issues that do not pass the <code>filter</code> condition. All issues
   * that contain sub-issues that have passed the filter are also preserved. The
   * resulting forest contains sub-sequence of the original forest with issues having the same parents and depths.</p>
   *
   * <p>Unlike {@link #filter} method, this method preserves the topology of the original forest -
   * all issues in the resulting forest have the same root path as they do in the original forest.</p>
   *
   * <p>This forest is not modified by this method. If all issues pass the condition, then this forest
   * is returned as the result. If filtering has taken place, a new forest is returned.</p>
   *
   * <p>The filter method is called once for every issue in the forest, and a truthy result
   * (as defined in {@link La#accepts}) indicates that the issue passes the filter.</p>
   *
   * @param filter filter that returns a truthy value if an issue with given ID should be present in the resulting forest
   * @return a filtered forest (or this forest if all issues satisfy the filter)
   */
  @NotNull
  Forest filterSoft(@Nullable La<Long, ?> filter);

  /**
   * <p>Creates a sub-sequence forest of this forest that contains all issues "relevant" to the passed issue.</p>
   *
   * <p>The resulting forest contains all the parent issues of <code>issue</code> up to the root issue, and
   * all the sub-issues of <code>issue</code> down to the deepest level.</p>
   *
   * <p>This filter preserves the topology of the forest - all issues in the resulting forest have the same parents
   * as in the original forest.</p>
   *
   * <p>If issue is not in this forest, returns empty forest.</p>
   *
   * @param issue the issue
   * @return a part of the forest that is relevant to the specified issue
   */
  @NotNull
  Forest filterForIssue(long issue);

  /**
   * Creates an exact copy of this forest.
   *
   * @return a copy of this forest
   */
  @NotNull
  Forest copy();

  /**
   * Creates a forest that contains the specified issue and all its sub-issues from this forest.
   *
   * @param issue the root of the sub-tree
   * @return a new forest that contains a copy of the sub-tree rooted at <code>issue</code>, or an empty forest if the issue is not in this forest
   */
  @NotNull
  Forest copySubtree(long issue);

  /**
   * Creates a forest that contains the specified issue and all its sub-issues from this forest.
   *
   * @param index index of the root of the sub-tree
   * @return a new forest that contains a copy of the sub-tree rooted at issue at <code>index</code>, or an empty forest if the index is negative
   */
  @NotNull
  Forest copySubtreeAtIndex(int index);


  // ======================== MUTATORS ========================

  /**
   * <p>Moves sub-tree rooted at the specified issue to a position specified by <code>(under, after)</code> coordinates.</p>
   *
   * <p>This method modifies the forest by removing the sub-tree with the specified issue as the root and adding it
   * at the position specified by <code>under-after</code> coordinates. See {@link Forest} for the explanation
   * of under-after coordinates.</p>
   *
   * <p><strong>This method does not modify the forest stored in the database for a structure.</strong>
   * To make modifications of the structure's forest, run forest changing transaction with {@link StructureManager#updateForest}.
   * </p>
   *
   * @param issue the root issue of the sub-tree being moved
   * @param under the new parent of the sub-tree, or 0 if the sub-tree should be placed at the forest root level
   * @param after the preceding sibling of the new location for the sub-tree, or 0 to place sub-tree as the first child
   * @param eventHandler optional handler of the move events
   * @return true if the sub-tree has been moved, false if not (for example, if the issue is not in the forest)
   * @throws StructureException if the move is not possible - for example, <code>under</code> is not in the forest or if you attempt to move a sub-tree under itself
   */
  boolean moveSubtree(long issue, long under, long after, @Nullable ForestChangeEventHandler eventHandler) throws StructureException;

  /**
   * Convenience method to call {@link #moveSubtree(long, long, long, ForestChangeEventHandler)} without event handler.
   *
   * @param issue the root issue of the sub-tree being moved
   * @param under the new parent of the sub-tree, or 0 if the sub-tree should be placed at the forest root level
   * @param after the preceding sibling of the new location for the sub-tree, or 0 to place sub-tree as the first child
   * @return true if the sub-tree has been moved, false if not (for example, if the issue is not in the forest)
   * @throws StructureException if the move is not possible - for example, <code>under</code> is not in the forest or if you attempt to move a sub-tree under itself
   */
  boolean moveSubtree(long issue, long under, long after) throws StructureException;

  /**
   * <p>Moves sub-tree rooted at the specified index to a position specified by <code>(under, after)</code> coordinates.</p>
   *
   * <p>This method modifies the forest by removing the sub-tree with the specified issue as the root and adding it
   * at the position specified by <code>under-after</code> coordinates. See {@link Forest} for the explanation
   * of under-after coordinates.</p>
   *
   * <p><strong>This method does not modify the forest stored in the database for a structure.</strong>
   * To make modifications of the structure's forest, run forest changing transaction with {@link StructureManager#updateForest}.
   * </p>
   *
   *
   * @param index the index of the root issue of the sub-tree being moved
   * @param under the new parent of the sub-tree, or 0 if the sub-tree should be placed at the forest root level
   * @param after the preceding sibling of the new location for the sub-tree, or 0 to place sub-tree as the first child
   * @param eventHandler optional handler of the move events
   * @return new index of the sub-tree if it has been moved, -1 if not (for example, if the given issue index is negative)
   * @throws StructureException if the move is not possible - for example, <code>under</code> is not in the forest or if you attempt to move a sub-tree under itself
   */
  int moveSubtreeAtIndex(int index, long under, long after, @Nullable ForestChangeEventHandler eventHandler) throws StructureException;

  /**
   * <p>Removes a sub-tree with the specified issue as a root from this forest.</p>
   *
   * <p><strong>This method does not modify the forest stored in the database for a structure.</strong>
   * To make modifications of the structure's forest, run forest changing transaction with {@link StructureManager#updateForest}.
   * </p>
   *
   * @param issue the root of the sub-tree to be removed
   * @param eventHandler optional handler of the forest events
   * @return forest with the removed issues, or empty forest if <code>issue</code> is not in this forest
   */
  @NotNull
  Forest removeSubtree(long issue, @Nullable ForestChangeEventHandler eventHandler);

  /**
   * Convenience method to call {@link #removeSubtree(long, ForestChangeEventHandler)} without event handler.
   *
   * @param issue the root of the sub-tree to be removed
   * @return forest with the removed issues, or empty forest if <code>issue</code> is not in this forest.
   */
  @NotNull
  Forest removeSubtree(long issue);

  /**
   * <p>Removes a sub-tree with rooted at the specified index from this forest.</p>
   *
   * <p><strong>This method does not modify the forest stored in the database for a structure.</strong>
   * To make modifications of the structure's forest, run forest changing transaction with {@link StructureManager#updateForest}.
   * </p>
   *
   * @param index the index of the root of the sub-tree to be removed
   * @param eventHandler optional handler of the forest events
   * @return forest with the removed issues, or empty forest if <code>index</code> is negative
   */
  @NotNull
  Forest removeSubtreeAtIndex(int index, @Nullable ForestChangeEventHandler eventHandler);

  /**
   * <p>Adds a single issue at the specified position.</p>
   *
   * <p>If the issue is already in the structure, this method moves it to the specified position.</p>
   *
   * <p><strong>This method does not modify the forest stored in the database for a structure.</strong>
   * To make modifications of the structure's forest, run forest changing transaction with {@link StructureManager#updateForest}.
   * </p>
   *
   * @param issue issue to be added
   * @param under the parent of the issue, or 0 if the issue should be placed at the forest root level
   * @param after the preceding sibling of the issue, or 0 to place issue as the first child
   * @return true if the forest has been changed
   * @throws StructureException if <code>under</code> is not in the forest or if a similar problem happens
   */
  boolean addIssue(long issue, long under, long after) throws StructureException;

  /**
   * <p>Merges another forest into this forest. After the method has executed, this forest will contain all issues
   * from <code>forest</code> in the same topology under the specified positions: the root issues from <code>forest</code>
   * will be positioned as defined by <code>under-after</code> coordinates (see {@link Forest} for the explanation
   * of under-after coordinates), and non-root issues from <code>forest</code> will have the same parent issues as
   * in <code>forest</code>.</p>
   *
   * <p>This forest and the merged <code>forest</code> may have the same issues. In that case the issues are moved
   * within this forest and placed at the position required by this operation.</p>
   *
   * <p><strong>This method does not modify the forest stored in the database for a structure.</strong>
   * To make modifications of the structure's forest, run forest changing transaction with {@link StructureManager#updateForest}.
   * </p>
   *
   * <p>
   *   When <code>StructureException</code> is thrown from this method, the state of this forest is unknown - due to the
   *   fact that merge splits into several atomic updates and the exception may be thrown after some of the updates
   *   have taken place.
   * </p>
   *
   * @param forest the merged forest
   * @param under the parent of the merged forest, or 0 if the forest issues should be placed at the root level
   * @param after the preceding sibling of the merged forest, or 0 to place forest as the first child
   * @param eventHandler an optional event handler to get notifications about changes being applied - may be called several times
   * because merge could be split into series of moves and additions.
   * @throws StructureException if the operation cannot be completed because it requires an invalid move
   */
  void mergeForest(@Nullable Forest forest, long under, long after, @Nullable ForestChangeEventHandler eventHandler)
    throws StructureException;


  /**
   * Convenience method to call {@link #mergeForest(Forest, long, long, ForestChangeEventHandler)} without event handler.
   *
   * @param forest the merged forest
   * @param under the parent of the merged forest, or 0 if the forest issues should be placed at the root level
   * @param after the preceding sibling of the merged forest, or 0 to place forest as the first child
   * @throws StructureException if the operation cannot be completed because it requires an invalid move
   */
  void mergeForest(@Nullable Forest forest, long under, long after) throws StructureException;

  /**
   * <p>Replaces the contents of this forest with the values passed in the parameters.</p>
   *
   * <p>The passed lists must satisfy the invariant conditions listed in {@link Forest}, otherwise the results
   * are undefined.</p>
   *
   * <p><strong>This method does not modify the forest stored in the database for a structure.</strong>
   * To make modifications of the structure's forest, run forest changing transaction with {@link StructureManager#updateForest}.
   * </p>
   *
   * @param issues new issue list
   * @param depths new depths list
   * @param reuseLists if true, the passed arrays may be used by the forest without copying the data. In that case, the calling code must not use these lists after this method call.
   * @return this forest
   */
  Forest set(WritableLongList issues, WritableIntList depths, boolean reuseLists);

  /**
   * <p>Replaces the contents of this forest with the values passed in the parameters.</p>
   *
   * <p>The passed lists must satisfy the invariant conditions listed in {@link Forest}, otherwise the results
   * are undefined.</p>
   *
   * <p><strong>This method does not modify the forest stored in the database for a structure.</strong>
   * To make modifications of the structure's forest, run forest changing transaction with {@link StructureManager#updateForest}.
   * </p>
   *
   * @param issues new issue list
   * @param depths new depths list
   * @return this forest
   */
  Forest set(LongList issues, IntList depths);


  // ======================== UTILITIES ========================

  /**
   * <p>This method is used to efficiently traverse all pairs (parent, children) from the end of the forest upwards.</p>
   *
   * <p>This method goes over the forest in the backwards direction and reports to the visitor
   * pairs of (parent, direct children).</p>
   * <p>Invariants:</p>
   * <ul>
   * <li>The number of calls to the visitor is equal to the forest size: every issue is reported as a parent once.</li>
   * <li>A child issue is reported (as the parent of its own sub-issues) before parent is reported:
   * the iteration goes upwards.</li>
   * <li>A leaf issue is reported as a parent with no children (empty children list).</li>
   * </ul>
   * <p>If the forest is modified during iteration, the results are undefined.</p>   *
   *
   * @param visitor an instance that receives the pairs of parent and children array
   */
  void visitParentChildrenUpwards(ForestParentChildrenVisitor visitor);

  /**
   * Makes this forest immutable - all mutator operations will throw <code>UnsupportedOperationException</code>.
   *
   * @return this forest
   */
  @NotNull
  Forest makeImmutable();

  /**
   * Utility method for debugging - returns full string representation of the forest, that contains all the information,
   * unlike <code>toString</code> method, which may be truncated to some character number limit.
   *
   * @return a full string containing all information about this forest
   */
  @NotNull
  String toFullString();
}
