package com.almworks.jira.structure.api.forest;

import com.almworks.jira.structure.api.StructureException;
import com.almworks.jira.structure.api.StructureManager;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * <p><code>ForestAccessor</code> is an interface that allows changing forests within
 * {@link ForestTransaction}. This is the only way to update structure contents.</p>
 *
 * <p>An instance of <code>ForestAccessor</code> is passed to an implementation of
 * <code>ForestTransaction</code> in a {@link ForestTransaction#transaction} method. The transaction
 * must use this accessor to make any modifications and to read current state of a forest.</p>
 *
 * <p><code>ForestAccessor</code> is short-lived: after transaction completes, the instance becomes
 * unusable, so it should not be stored anywhere.</p>
 *
 * <p>All methods are thread-unsafe. Use <code>ForestAccessor</code> only within the thread and stack
 * of the <code>ForestTransaction.transaction</code> method.</p>
 *
 * <p><code>ForestAccessor</code> methods execute under the credentials of the user specified
 * in the call to <code>StructureManager.updateForest</code> method.</p>
 *
 * @author Igor Sereda
 */
public interface ForestAccessor {
  /**
   * Returns the current forest for the given structure. The returned forest accounts for all changes
   * made to this structure in the current transaction - unlike {@link StructureManager#getForest},
   * which always returns the persisted state of a forest.
   *
   * @param structureId the ID of the structure
   * @return current structure's forest
   * @throws StructureException if there's no such structure or if permission was denied
   */
  @NotNull
  Forest getForest(@Nullable Long structureId) throws StructureException;

  /**
   * <p>Moves sub-tree rooted at the specified issue to a position specified by <code>(under, after)</code> coordinates.</p>
   *
   * <p>This method modifies the forest by removing the sub-tree with the specified issue as the root and adding it
   * at the position specified by <code>under-after</code> coordinates. See {@link Forest} for the explanation
   * of under-after coordinates.</p>
   *
   * @param structure structure ID
   * @param issue the root issue of the sub-tree being moved
   * @param under the new parent of the sub-tree, or 0 if the sub-tree should be placed at the forest root level
   * @param after the preceding sibling of the new location for the sub-tree, or 0 to place sub-tree as the first child
   * @return true if the sub-tree has been moved, false if not (for example, if the issue is not in the forest)
   * @throws StructureException if structure is not found or permissions or other problem has happened
   */
  boolean moveSubtree(@Nullable Long structure, @Nullable Long issue, @Nullable Long under,
    @Nullable Long after) throws StructureException;

  /**
   * <p>Adds a single issue at the specified position. If the issue is already in the structure, this method moves it to the specified position.</p>
   *
   * <p>See {@link Forest} for the explanation of under-after coordinates.</p>
   *
   * @param structure structure ID
   * @param issue issue to be added
   * @param under the parent of the issue, or 0 if the issue should be placed at the forest root level
   * @param after the preceding sibling of the issue, or 0 to place issue as the first child
   * @return true if the forest has been changed
   * @throws StructureException if structure is not found or permissions or other problem has happened
   */
  boolean addIssue(@Nullable Long structure, @Nullable Long issue, @Nullable Long under,
    @Nullable Long after) throws StructureException;

  /**
   * <p>Removes a sub-tree with the specified issue as a root from this forest.</p>
   *
   * @param structure structure ID
   * @param issue the root of the sub-tree to be removed
   * @return forest with the removed issues, or empty forest if <code>issue</code> is not in this forest
   * @throws StructureException if structure is not found or permissions or other problem has happened
   */
  @NotNull
  Forest removeSubtree(@Nullable Long structure, @Nullable Long issue) throws StructureException;

  /**
   * <p>Merges another forest into the structure's forest. After the method has executed, the structure forest will contain all issues
   * from <code>forest</code> in the same topology under the specified positions: the root issues from <code>forest</code>
   * will be positioned as defined by <code>under-after</code> coordinates (see {@link Forest} for the explanation
   * of under-after coordinates), and non-root issues from <code>forest</code> will have the same parent issues as
   * in <code>forest</code>.</p>
   *
   * <p>This forest and the merged <code>forest</code> may have the same issues. In that case the issues are moved
   * within this forest and placed at the position required by this operation.</p>
   *
   * <p>
   * When <code>StructureException</code> is thrown from this method, the state of this forest is unknown - due to the
   * fact that merge splits into several atomic updates and the exception may be thrown after some of the updates
   * have taken place.
   * </p>
   *
   * @param structureId structure ID
   * @param forest the merged forest
   * @param under the parent of the merged forest, or 0 if the forest issues should be placed at the root level
   * @param after the preceding sibling of the merged forest, or 0 to place forest as the first child
   * @throws StructureException if structure is not found or permissions or other problem has happened
   */
  void mergeForest(@Nullable Long structureId, @Nullable Forest forest, @Nullable Long under,
    @Nullable Long after) throws StructureException;
}
