package com.almworks.jira.structure.api.forest;

import com.almworks.jira.structure.api.StructureException;
import com.almworks.jira.structure.api.StructureManager;
import org.jetbrains.annotations.NotNull;

/**
 * <p>Implement this interface to define a single transaction on the forests.</p>
 *
 * <p>
 *   <strong>It is very important to implement <code>ForestTransaction</code> correctly.</strong> The transaction
 *   may be executed under one or more write locks, preventing other transactions from changing the forests
 *   concurrently (readers still have access) - so the amount of code should be minimal and the execution time
 *   should be predictable and small.
 * </p>
 *
 * @param <T> type of the operation result, use <code>Void</code> or <code>Object</code>
 * when no result needed
 * @see StructureManager#updateForest
 */
public interface ForestTransaction<T> {
  /**
   * <p>Perform an operation on the forests in the structure database.</p>
   *
   * <p>If this method throws any exception, the whole transaction will be considered failed and changes
   * will be rolled back.</p>
   *
   * @param accessor accessor that is used to read forest and execute updates
   * @return optional result that will be passed to the calling code as the result of {@link StructureManager#updateForest}.
   * @throws StructureException when operation fails
   */
  T transaction(@NotNull ForestAccessor accessor) throws StructureException;
}
