package com.almworks.jira.structure.api.forest;

import org.jetbrains.annotations.NotNull;

import java.util.Collections;
import java.util.List;

/**
 * <p><code>ForestUpdate</code> represents information about what the client code needs to do
 * to bring their version of some structure's forest to the some newer version.</p>
 *
 * <p>There are three classes of the forest update, and there cannot be any others:</p>
 * <ul>
 *   <li>{@link ForestUpdate.Empty} means that the client has up-to-date version of the forest.</li>
 *   <li>{@link ForestUpdate.Incremental} means that the client can bring their version of the forest up to date by applying a list of forest transformations.</li>
 *   <li>{@link ForestUpdate.Full} contains the full version of the forest.</li>
 * </ul>
 *
 * <p>
 *   The instances of this class are created in <code>getForestUpdate</code> method, generally there's no need to
 *   create instances yourself.
 * </p>
 *
 * @see com.almworks.jira.structure.api.StructureManager#getForestUpdate
 * @author Igor Sereda
 */
public abstract class ForestUpdate {
  private final long myStructureId;
  private final long myVersion;

  protected ForestUpdate(long structureId, long version) {
    myStructureId = structureId;
    myVersion = version;
  }

  /**
   * @return the ID of the structure this update is about
   */
  public long getStructureId() {
    return myStructureId;
  }

  /**
   * @return the version of the structure's forest - after applying this update the client code will have
   * this version of the forest
   */
  public long getVersion() {
    return myVersion;
  }


  /**
   * Empty update means that nothing needs to be done and the client code has the latest version of the forest.
   */
  public static class Empty extends ForestUpdate {
    public Empty(long structureId, long version) {
      super(structureId, version);
    }
  }


  /**
   * Incremental update means that the client code should apply a series of transformations
   * to bring their version of the forest to the latest version.
   */
  public static class Incremental extends ForestUpdate {
    private final long myFromVersion;
    private final List<ForestOp> myUpdates;

    public Incremental(long structureId, long lastVersion, long fromVersion, List<ForestOp> updates) {
      super(structureId, lastVersion);
      myFromVersion = fromVersion;
      myUpdates = Collections.unmodifiableList(updates);
    }

    /**
     * @return the version of the forest that this update is for. Before applying the operations,
     * the client code should ensure that they have exactly this version of the forest.
     */
    public long getFromVersion() {
      return myFromVersion;
    }

    /**
     * @return a read-only list of the polymorphic updates for the forest - essentially, a diff from forest
     * at version {@link #getFromVersion()} and forest at version {@link #getVersion()}
     */
    @NotNull
    public List<ForestOp> getUpdates() {
      return myUpdates;
    }
  }


  /**
   * The full update contains the whole forest. The client code should throw away the old version of the forest
   * and use the new one.
   */
  public static class Full extends ForestUpdate {
    private final Forest myForest;

    public Full(long structureId, long version, Forest forest) {
      super(structureId, version);
      myForest = forest;
    }

    /**
     * @return the structure's forest at version {@link #getVersion()}
     */
    @NotNull
    public Forest getForest() {
      return myForest;
    }
  }
}
