package com.almworks.jira.structure.api.progress;

import com.almworks.integers.LongList;
import com.almworks.jira.structure.api.StructureException;
import com.almworks.jira.structure.api.forest.Forest;
import com.atlassian.jira.issue.Issue;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * <p><code>ProgressCalculator</code> is used to calculate aggregate progress of an issue
 * in a forest.</p>
 *
 * <p><code>ProgressCalculator</code> also maintains a cache of pre-calculated progress information for
 * issues in the structures. </p>
 *
 * @see <a href="http://wiki.almworks.com/display/structure/Progress+Column">Progress Column</a> for the explanation about how progress is calculated
 */
public interface ProgressCalculator {
  /**
   * <p>Returns progress info for the given issue in an arbitrary forest.
   * If either issue or forest parameters are null, returns empty progress.
   * If the issue is not in the forest, returns empty progress.</p>
   *
   * <p>This method does not use any caches.</p>
   *
   * @param issue the issue to calculate the progress for
   * @param forest the forest
   * @return progress info
   */
  @NotNull
  ProgressInfo getProgressInfo(@Nullable Issue issue, @Nullable Forest forest);

  /**
   * <p>Returns progress info for a specific issue in the given structure. This method uses caching so it can
   * be considerably faster than {@link #getProgressInfo(com.atlassian.jira.issue.Issue, com.almworks.jira.structure.api.forest.Forest)}.</p>
   *
   * <p>This method does not check structure access, so the calling code must ensure that the user has access
   * to this structure and this issue. The calculated progress accounts for all issues in the structure,
   * regardless of the user's permissions.</p>
   *
   * @param structureId the structure ID
   * @param issue issue
   * @return progress info
   * @throws StructureException if there's no structure with the given ID
   */
  @NotNull
  ProgressInfo getProgressInfo(@Nullable Long structureId, @Nullable Issue issue) throws StructureException;

  /**
   * <p>Calculates progress for a collection of issues in the specified structure.
   * This method uses caches maintained by the progress calculator, and it is also more efficient than calling
   * {@link #getProgressInfo(Long, com.atlassian.jira.issue.Issue)} for each issue separately.</p>
   *
   * <p>The returned map contains non-null <code>ProgressInfo</code> for every issue requested.
   * In addition, the returned map may contain progress calculated for other issues in the forest, but don't
   * count on that.</p>
   *
   * <p>This method does not check structure access, so the calling code must ensure that the user has access
   * to this structure and this issue. The calculated progress accounts for all issues in the structure,
   * regardless of the user's permissions.</p>
   *
   * @param structureId the structure ID
   * @param issues a collection of issues to have progress calculated
   * @return a map from issue ID to the calculated progress info
   * @throws StructureException if there's no structure with the specified ID
   */
  @NotNull
  Map<Long, ProgressInfo> getProgressInfoMap(@Nullable Long structureId, @Nullable Collection<? extends Issue> issues)
    throws StructureException;

  /**
   * Clears internal cache of pre-calculated progress info. Normally it's not required because <code>ProgressCalculator</code>
   * keeps track of issue changes and structure changes and invalidates parts of the cache.
   */
  void clearCache();
}
