package com.almworks.jira.structure.api.progress;

import java.util.Locale;

/**
 * <p><code>ProgressInfo</code> represents aggregate information about time and progress,
 * calculated for a sub-tree of issues.</p>
 */
public class ProgressInfo {
  public static final ProgressInfo ZERO = new ProgressInfo(0, 0, 0, 0.0);

  private final long myOriginalEstimate;
  private final long myCurrentEstimate;
  private final long myTimeSpent;

  // calculated based on children progress, if there are any
  private final double myProgress;

  /**
   * Creates an instance of progress info.
   *
   * @param originalEstimate total original estimate, in milliseconds
   * @param currentEstimate total current estimate, in milliseconds
   * @param timeSpent total time spent, in milliseconds
   * @param progress the progress, from 0.0 to 1.0
   */
  public ProgressInfo(long originalEstimate, long currentEstimate, long timeSpent, double progress) {
    assert originalEstimate >= 0 : originalEstimate;
    assert currentEstimate >= 0 : currentEstimate;
    assert timeSpent >= 0 : timeSpent;
    assert progress >= 0 && progress <= 1 : progress;
    myOriginalEstimate = originalEstimate;
    myCurrentEstimate = currentEstimate;
    myTimeSpent = timeSpent;
    myProgress = progress;
  }

  public String toString() {
    return "ProgressInfo(" +
      "oe=" + myOriginalEstimate +
      ", ce=" + myCurrentEstimate +
      ", ts=" + myTimeSpent +
      ", p=" + String.format(Locale.US, "%.2f", myProgress) +
      ')';
  }

  /**
   * @return true if the progress does not contain any data (progress is 0 and no time or estimation is there)
   */
  public boolean isEmpty() {
    return myOriginalEstimate == 0 && myCurrentEstimate == 0 && myTimeSpent == 0 && myProgress < 1e-5;
  }

  /**
   * @return true if the progress contains any timing information
   */
  public boolean hasTime() {
    return myOriginalEstimate > 0 || myCurrentEstimate > 0 || myTimeSpent > 0;
  }

  /**
   * @return total original estimate, in milliseconds
   */
  public long getOriginalEstimate() {
    return myOriginalEstimate;
  }

  /**
   * @return total current estimate, in milliseconds
   */
  public long getCurrentEstimate() {
    return myCurrentEstimate;
  }

  /**
   * @return total time spent, in milliseconds
   */
  public long getTimeSpent() {
    return myTimeSpent;
  }

  /**
   * @return progress, value from 0.0 to 1.0
   */
  public double getProgress() {
    return myProgress;
  }

  /**
   * @return a string representing the progress, like "23%"
   */
  public String getProgressString() {
    double p = getProgress();
    return (int) Math.round(100.0 * p) + "%";
  }

  public boolean equals(Object o) {
    if (this == o) return true;
    if (o == null || getClass() != o.getClass()) return false;

    ProgressInfo that = (ProgressInfo) o;

    if (myCurrentEstimate != that.myCurrentEstimate) return false;
    if (myOriginalEstimate != that.myOriginalEstimate) return false;
    if (myTimeSpent != that.myTimeSpent) return false;
    if (Math.abs(myProgress - that.myProgress) > 1e-5) return false;

    return true;
  }

  public int hashCode() {
    int result;
    long temp;
    result = (int) (myOriginalEstimate ^ (myOriginalEstimate >>> 32));
    result = 31 * result + (int) (myCurrentEstimate ^ (myCurrentEstimate >>> 32));
    result = 31 * result + (int) (myTimeSpent ^ (myTimeSpent >>> 32));
    temp = Math.round(myProgress * 1000);
    result = 31 * result + (int) (temp ^ (temp >>> 32));
    return result;
  }
}
