package com.almworks.jira.structure.api.sync;

import com.almworks.jira.structure.api.StructureServices;
import com.almworks.jira.structure.api.event.IssueEventBridge;
import com.almworks.jira.structure.api.event.IssueListener;
import org.jetbrains.annotations.NotNull;

import java.util.concurrent.atomic.AtomicReference;

/**
 * <p>This is the base class for the synchronizer that need to listen to issue changes in JIRA.</p>
 *
 * <p><code>AbstractIssueListeningSynchronizer</code> subscribes to the events provided by
 * {@link IssueEventBridge}</p> and requests synchronization when an event comes.</p>
 *
 * @author Igor Sereda
 */
public abstract class AbstractIssueListeningSynchronizer<P> extends AbstractSynchronizer<P> {
  protected static final String LISTENER_KEY = "listener";

  protected final IssueEventBridge myEventBridge;

  /**
   * Constructs an instance
   *
   * @param services services facade
   * @param parametersClass the class of the parameters
   */
  protected AbstractIssueListeningSynchronizer(StructureServices services, Class<P> parametersClass) {
    super(services, parametersClass);
    myEventBridge = services.getIssueEventBridge();
  }

  /**
   * Starts listening to issue events. This method is called by the sync manager, you don't need
   * to call it. See {@link StructureSynchronizer#start}. You can override it if you need
   * additional listening to be started.
   *
   * @param instance sync instance
   * @param controller controller
   */
  public void start(@NotNull SyncInstance instance, @NotNull SyncController controller) {
    AtomicReference<BasicListener> ref = instance.transientData(LISTENER_KEY);
    if (ref.get() != null) return;
    BasicListener listener = createIssueListener(instance, controller);
    if (listener == null) return;
    if (ref.compareAndSet(null, listener)) {
      myEventBridge.addListener(listener);
    }
  }

  /**
   * Creates an instance of basic listener for the given sync instance and controller.
   * Override this method to provide your extension of the basic listener.
   *
   * @param instance sync instance
   * @param controller sync controller
   * @return the listener
   */
  @NotNull
  protected BasicListener createIssueListener(@NotNull SyncInstance instance, @NotNull SyncController controller) {
    return new BasicListener(instance, controller);
  }

  /**
   * Stops listening issue events. This method is called by the sync manager, you don't need
   * to call it. See {@link StructureSynchronizer#stop}. You can override it if you need
   * additional listening to be stopped.
   *
   * @param instance sync instance
   */
  public void stop(@NotNull SyncInstance instance) {
    AtomicReference<BasicListener> ref = instance.transientData(LISTENER_KEY);
    BasicListener listener = ref.get();
    if (listener == null) return;
    if (ref.compareAndSet(listener, null)) {
      listener.stop();
      myEventBridge.removeListener(listener);
    }
  }

  /**
   * <p>Lets the synchronizer decide whether the issue should be reported as requiring synchronization.</p>
   *
   * <p>For example, if synchronizer configuration limits the scope to only a single project,
   * you can override this method and inspect if the issue belongs to that project.
   * This is for optimization only - you still need to check that condition when running
   * sync.</p>
   *
   * @param instance sync instance
   * @param issueId the ID of the issue
   * @return true if the issue should be reported as requiring synchronization
   */
  protected boolean isIssueChangeInteresting(SyncInstance instance, long issueId) {
    return true;
  }


  /**
   * <p>The basic implementation of issue listener that reports updated issues
   * as requiring synchronization if the following is true:</p>
   * <ol>
   *   <li>This synchronizer is not already running - then the change may be coming from it.</li>
   *   <li>The {@link AbstractIssueListeningSynchronizer#isIssueChangeInteresting} method returns <code>true</code>.</li>
   * </ol>
   *
   */
  protected class BasicListener implements IssueListener {
    private final SyncInstance myInstance;
    private final SyncController myController;
    private volatile boolean myStopped;

    public BasicListener(SyncInstance instance, SyncController controller) {
      myInstance = instance;
      myController = controller;
    }

    public void onIssueChanged(long issueId) {
      if (myStopped) return;
      if (isResyncRequired(issueId)) {
        myController.incrementalSyncRequired(issueId);
      }
    }

    protected boolean isResyncRequired(long issueId) {
      return !myController.isThisSyncRunningInCurrentThread() && isIssueChangeInteresting(myInstance, issueId);
    }

    public void stop() {
      myStopped = true;
    }
  }
}
