package com.almworks.jira.structure.api.sync;

/**
 * @author Igor Sereda
 */

import com.almworks.jira.structure.api.job.StructureJobManager;
import com.almworks.jira.structure.util.StructureUtil;
import com.atlassian.jira.web.action.JiraWebActionSupport;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.util.List;
import java.util.Map;

/**
 * <p>A <code>StructureSynchronizer</code> is a pluggable Structure extension that
 * allows you to sync structure with any other aspect of issues, be it links, subtasks or anything
 * else. By implementing this interface and registering it with <code>&lt;structure-synchronizer&gt;</code>
 * module, you define a new type of the synchronizer, which can be accessed from
 * the Structure's synchronization features.</p>
 *
 * <p>To get started with synchronizers, read <a href="http://wiki.almworks.com/display/structure/Synchronization">section about Synchronization in the Structure user documentation</a>.</p>
 *
 * <p>
 *   You typically implement a synchronizer by extending {@link AbstractSynchronizer}
 *   or {@link AbstractIssueListeningSynchronizer}.
 * </p>
 *
 * <p>Two main methods - {@link #resync} and {@link #sync} - do the main synchronization job,
 * while other methods are used to create synchronizer configuration and provide necessary information
 * for the Structure plugin.</p>
 *
 * <p>The instance of <code>StructureSynchronizer</code> represents a <i>type of synchronization</i> and
 * is instantiated only once. The synchronizer instances that the users install into a structure
 * are represented by {@link SyncInstance} interface.</p>
 *
 * <p>Synchronizer is configured with some kind of parameters, which are serialized via
 * {@link #storeParameters(Object)} and {@link #restoreParameters(byte[])} and stored in
 * the Structure database. The type of parameters is up to the synchronizer, whenever parameters
 * are passed they have type <code>Object</code>, so the synchronizer needs to cast them to its
 * parameters type.</p>
 */
public interface StructureSynchronizer {

  /**
   * <p>Checks if this type of synchronizer is currently available. A synchronizer may not be available
   * if, for example, some feature is turned off; for example, subtasks synchronizer would be
   * unavailable if sub-tasks are turned off in JIRA.</p>
   *
   * <p>This method is only checked when the user selects from synchronizers and when synchronizer
   * is being installed.</p>
   *
   * <p>The method is called under a lock and should return as fast as possible. The implementation
   * must not use methods of {@link StructureSyncManager} directly or indirectly.</p>
   *
   * @return true if the synchronizer can be used
   */
  boolean isAvailable();

  /**
   * <p>Checks if synchronizer supports automatic incremental synchronization.</p>
   *
   * <p>If autosync is supported, the synchronizer may be installed and enabled for a structure.
   * If it's not supported, the synchronizer may only be used to run resync (or import/export).</p>
   *
   * <p>The method is called under a lock and should return as fast as possible. The implementation
   * must not use methods of {@link StructureSyncManager} directly or indirectly.</p>
   *
   * @return true if the synchronizer supports autosync with {@link #sync} method
   */
  boolean isAutosyncSupported();

  /**
   * <p>Checks if the resync (import/export) is supported for the given direction.</p>
   *
   * @param direction the direction of resync ({@link SyncDirection#INBOUND} for Import, {@link SyncDirection#OUTBOUND} for Export).
   * @return true if the synchronizer can resync in the given direction
   */
  boolean isDirectionSupported(@NotNull SyncDirection direction);

  /**
   * <p>Creates a short one-line description of the configuration parameters, shown in
   * a few places in the Structure interface, such as on the Manage Structure page.</p>
   *
   * @param parameters sync parameters
   * @return a string describing the configuration, or null
   */
  @Nullable
  String getConfigDescription(@Nullable Object parameters);

  /**
   * Creates a list of strings that fully describe the synchronizer's configuration. Used on some pages such as
   * Synchronization Settings.
   *
   * @param parameters sync parameters
   * @return a list of lines, describing the configuration, or null
   */
  @Nullable
  List<String> getConfigDescriptionDetails(@Nullable Object parameters);

  /**
   * <p>Converts parameters into array of bytes for storing them in the database. The result of using
   * {@link #restoreParameters(byte[])} on the resulting byte array should reconstruct the same parameters object.</p>
   *
   * <p>The method is called under a lock and should return as fast as possible. The implementation
   * must not use methods of {@link StructureSyncManager} directly or indirectly.</p>
   *
   * @param parameters sync parameters
   * @return byte array with the result
   * @throws IOException if parameters cannot be stored
   */
  @Nullable
  byte[] storeParameters(@Nullable Object parameters) throws IOException;

  /**
   * Converts byte array previously stored with {@link #storeParameters(Object)} into this synchronizer's
   * parameters object.
   *
   * @param data byte array with serialized parameters
   * @return the parameters object
   * @throws IOException if there's a problem reading parameters
   */
  @Nullable
  Object restoreParameters(@Nullable byte[] data) throws IOException;

  /**
   * @return module descriptor, which is used to get configuration properties from the plugin XML descriptor
   */
  @NotNull
  StructureSynchronizerModuleDescriptor getDescriptor();

  /**
   * <p>Creates an instance of synchronizer parameters. If parameters cannot be created, the method
   * should return null and add errors to the action.</p>
   *
   * <p>To read the parameters, you can use {@link StructureUtil#getSingleParameter}
   * method and others like it.</p>
   *
   * @param formParameters the map from <code>String</code> to values that is constructed from the HTML form
   * parameters sent by the browser
   * @param action the action executing the update - use it to report errors
   * @return the created parameters, or null if they cannot be created
   * @see JiraWebActionSupport#addError
   * @see JiraWebActionSupport#addErrorMessage
   * @see StructureUtil#getSingleParameter
   * @see StructureUtil#getSingleParameterLong
   * @see StructureUtil#getSingleParameterBoolean
   */
  @Nullable
  Object buildParametersFromForm(@NotNull Map<String, ?> formParameters, @NotNull JiraWebActionSupport action);

  /**
   * Adds to the map the default values for the parameters in the synchronizer parameters form.
   * The default values then will be available for the synchronizer's &lt;form&gt; the first time
   * it's loaded.
   *
   * @param params the map of parameters that the synchronizer can add to
   */
  void addDefaultFormParameters(@NotNull Map<String, Object> params);

  /**
   * <p>Perform full resync in the specified direction.</p>
   *
   * <p>This method is called when the user request full resync or runs Import or Export.</p>
   *
   * <p>The implementation of this method should make structure and other aspect of an issue synchronized,
   * inspecting and making changes to all issues that are subject for synchronization according to the
   * synchronizer's configuration.</p>
   *
   * @param instance the configured instance of the synchronizer
   * @param direction the direction of the resync
   * @see StructureSyncManager#resync
   */
  void resync(@NotNull SyncInstance instance, @NotNull SyncDirection direction);

  /**
   * <p>Perform incremental synchronization.</p>
   *
   * <p>This method is called when the synchronizer is installed and enabled, and sync manager detects changes
   * in JIRA issues or in the structure. The lists of changed issues is passed via <code>data</code> parameter
   * and allows the synchronizer to check only those issues that have been affected and also to choose the
   * direction of the synchronization based on where the changes have occurred.</p>
   *
   * <p>The process is decoupled from the changing thread; the synchronization is run as a separate background
   * job with {@link StructureJobManager} shortly after the changes have taken place.</p>
   *
   * @param instance the configured instance of the synchronizer
   * @param data changed data
   */
  void sync(@NotNull SyncInstance instance, @NotNull IncrementalSyncData data);

  /**
   * <p>This method is called when a synchronizer instance becomes enabled. It allows the synchronizer to subscribe
   * to external events to be able to request incremental synchronization when issue data changes.</p>
   *
   * <p>The generic {@link AbstractIssueListeningSynchronizer} uses this method to start listening to JIRA issue
   * events.</p>
   *
   * @param instance the configured instance of the synchronizer
   * @param controller synchronization controller used to request synchronization
   */
  void start(@NotNull SyncInstance instance, @NotNull SyncController controller);

  /**
   * This method is called when a synchronizer instance becomes disabled. The synchronizer should stop listening
   * to any other external events and relinquish the saved instance of the {@link SyncController}.
   *
   * @param instance the configured instance of the synchronizer
   */
  void stop(@NotNull SyncInstance instance);
}
