package com.almworks.jira.structure.api.sync;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.concurrent.atomic.AtomicReference;

/**
 * <p><code>SyncInstance</code> represents an installed and configured instance of a synchronizer.
 * It provides the synchronizer with access to parameters that the user has selected
 * and access to instance-specific caches.</p>
 *
 * <p>The synchronizer receives <code>SyncInstance</code> parameter whenever it is asked
 * to do an instance-specific task such as syncing.</p>
 *
 * @author Igor Sereda
 */
public interface SyncInstance {
  /**
   * @return the ID of this synchronizer instance, or <code>0</code> if the synchronizer
   * is not installed (for example, when executed with {@link StructureSyncManager#resyncOnce}).
   */
  long getInstanceId();

  /**
   * @return the synchronizer for this sync instance
   */
  @NotNull
  StructureSynchronizer getSynchronizer();

  /**
   * @return the ID of the structure this sync instance is installed on
   */
  long getStructureId();

  /**
   * Synchronizer's parameters. The type of the object is controlled by the
   * synchronizer's {@link StructureSynchronizer#buildParametersFromForm}
   * and {@link StructureSynchronizer#restoreParameters} methods.
   *
   * @return the parameters of the synchronization
   */
  @Nullable
  Object getParameters();

  /**
   * @return user name of the user account that is used to run synchronizer jobs with this sync instance
   */
  @Nullable
  String getUsername();

  /**
   * Provides the synchronizer with a temporary / cache storage. The <code>key</code> parameter is an arbitrary
   * string that the synchronizer can use to have storage for several types of the information. The values are
   * stored for this sync instance only.
   *
   * @param key data key
   * @param <T> type of data stored
   * @return the container for the stored data
   */
  @NotNull
  <T> AtomicReference<T> transientData(String key);
}

