package com.almworks.jira.structure.util;

import com.atlassian.jira.util.Predicate;
import com.google.common.base.Function;

import java.lang.reflect.Array;
import java.util.*;

/**
 * <p><code>La</code> is a utility generic-purpose class for functional expressions.</p>
 *
 * @author Igor Sereda
 */
public abstract class La<T, R> implements Function<T, R>, Predicate<T> {
  private static final La SELF = new La() {
    public Object la(Object argument) {
      return argument;
    }
  };
  private static final La TRUE = new La() {
    public Object la(Object argument) {
      return Boolean.TRUE;
    }
  };
  private static final La FALSE = new La() {
    public Object la(Object argument) {
      return Boolean.FALSE;
    }
  };

  private final Class<R> rClass;

  protected La() {
    rClass = null;
  }

  /**
   * Use this constructor to have class-aware instance, which is needed only for
   * few methods like {@link #array}.
   *
   * @param rClass the class of the result
   */
  protected La(Class<R> rClass) {
    this.rClass = rClass;
  }

  public abstract R la(T argument);

  // ============== constructors ================

  public static <T, R> La<T, R> adapt(final Function<T, R> f) {
    return new La<T, R>() {
      @Override
      public R la(T argument) {
        return f.apply(argument);
      }
    };
  }

  public static <T, R> La<T, R> constant(final R result) {
    if (result == Boolean.TRUE) return (La<T, R>) TRUE;
    if (result == Boolean.FALSE) return (La<T, R>) FALSE;
    return new La<T, R>() {
      public R la(T argument) {
        return result;
      }
    };
  }

  public static <T> La<T, Boolean> isEqual(final T value) {
    final boolean isNull = value == null;
    return new La<T, Boolean>() {
      public Boolean la(T argument) {
        return isNull ? argument == null : value.equals(argument);
      }
    };
  }

  // we could declare the parameter as Collection<? super T>, but then we'd lose type inference for the calling line
  public static <T> La<T, Boolean> inCollection(final Collection<T> collection) {
    if (collection == null) return constant(false);
    return new La<T, Boolean>() {
      public Boolean la(T argument) {
        return collection.contains(argument);
      }
    };
  }
  
  public static <T> La<T, Boolean> instanceOf(final Class<? extends T> instanceClass) {
    if (instanceClass == null) return constant(false);
    return new La<T, Boolean>() {
      public Boolean la(T argument) {
        return instanceClass.isInstance(argument);
      }
    };
  }

  public static <A, B, C> La<A, C> compose(final La<B, C> g, final La<A, ? extends B> f) {
    return new La<A, C>() {
      public C la(A argument) {
        return g.la(f.la(argument));
      }
    };
  }

  public static <T> La<T, T> self() {
    return (La<T, T>) SELF;
  }

  // ============== utilities ================

  @Override
  public final R apply(T from) {
    return la(from);
  }

  @Override
  public boolean evaluate(T input) {
    return accepts(input);
  }

  public <A> La<A, R> apply(final La<A, ? extends T> f) {
    return La.compose(this, f); // why highlight?
  }

  public <A> La<T, A> supply(final La<? super R, A> g) {
    return La.compose(g, this); // why highlight?
  }

  public <D extends T> List<D> filter(Collection<D> collection) {
    ArrayList<D> r = new ArrayList<D>();
    if (collection != null) {
      for (D item : collection) {
        if (accepts(item)) {
          r.add(item);
        }
      }
    }
    return r;
  }

  public boolean accepts(T value) {
    return acceptsResult(la(value));
  }

  private boolean acceptsResult(R value) {
    return value != null && !value.equals(false) && !value.equals(0) && !value.equals(0L);
  }

  public Set<R> hashSet(Collection<? extends T> from) {
    return addTo(from, new HashSet<R>(from.size()), true);
  }

  public List<R> arrayList(Collection<? extends T> from) {
    return arrayList(from, true);
  }

  public List<R> arrayList(Collection<? extends T> from, boolean acceptFalsy) {
    return addTo(from, new ArrayList<R>(from.size()), acceptFalsy);
  }

  public <C extends Collection<R>> C addTo(Collection<? extends T> from, C to, boolean acceptFalsy) {
    for (T t : from) {
      R r = la(t);
      if (acceptFalsy || acceptsResult(r)) {
        to.add(r);
      }
    }
    return to;
  }

  public R[] array(Collection<? extends T> from) {
    Class<R> cls = rClass;
    if (cls == null) { throw new UnsupportedOperationException("cannot create array without rClass (" + from + ")"); }
    R[] r = (R[]) Array.newInstance(cls, from.size());
    int i = 0;
    for (T t : from) {
      if (i >= r.length) {
        throw new IllegalStateException("collection changed on the go");
      }
      r[i++] = la(t);
    }
    return r;
  }

  public Map<R, T> mapInto(Collection<? extends T> fromCollection, Map<R, T> toMap) {
    return mapInto(fromCollection, toMap, La.<T>self());
  }

  public <V> Map<R, V> mapInto(Collection<? extends T> fromCollection, Map<R, V> toMap, La<T, V> valueFunction) {
    if (fromCollection != null) {
      for (T value : fromCollection) {
        R key = la(value);
        if (key != null) {
          toMap.put(key, valueFunction.la(value));
        }
      }
    }
    return toMap;
  }

  public Map<R, T> hashMap(Collection<? extends T> collection) {
    return hashMap(collection, La.<T>self());
  }

  public <V> Map<R, V> hashMap(Collection<? extends T> collection, La<T, V> valueFunction) {
    return mapInto(collection, new HashMap<R, V>(collection.size()), valueFunction);
  }

  public <V> Map<R, V> linkedHashMap(Collection<? extends T> collection, La<T, V> valueFunction) {
    return mapInto(collection, new LinkedHashMap<R, V>(collection.size()), valueFunction);
  }

  public int indexOf(List<? extends T> list, R sample) {
    if (list == null) return -1;
    for (int i = 0; i < list.size(); i++) {
      R v = la(list.get(i));
      if (sample == null && v == null || sample != null && sample.equals(v)) return i;
    }
    return -1;
  }

  public <X extends T> X find(Collection<X> collection, R sample) {
    if (collection == null || collection.isEmpty()) return null;
    for (X value : collection) {
      R r = la(value);
      if (r == null && sample == null) {
        return value;
      } else if (r != null && r.equals(sample)) {
        return value;
      }
    }
    return null;
  }

  public La<T, Boolean> not() {
    final La<T, R> positive = this;
    return new La<T, Boolean>() {
      public Boolean la(T argument) {
        return !positive.accepts(argument);
      }
    };
  }
}

